"""Test the species defaults"""

from pyfhiaims.species_defaults.species import SpeciesDefaults

species_content_in = """################################################################################
#
#  FHI-aims code project
#  Volker Blum, Fritz Haber Institute Berlin, 2009
#
#  Suggested "tight" defaults for Si atom (to be pasted into control.in file)
#
#  Revised Jan 04, 2011, following tests (SiC) done by Lydia Nemec:
#     d and g functions of tier 2 now enabled by default.
#
################################################################################

  species        Si
#     global species definitions
    nucleus             14
    mass                28.0855
#
    l_hartree           6
#
    cut_pot             4.0          2.0  1.0
    basis_dep_cutoff    1e-3
#
    radial_base         42 7.0
    radial_multiplier   2
    angular_grids       specified
      division   0.4121   50
      division   0.7665  110
      division   1.0603  194
      division   1.2846  302
      division   1.4125  434
#      division   1.4810  590
#      division   1.5529  770
#      division   1.6284  974
#      division   2.6016 1202
#      outer_grid   974
      outer_grid  434
################################################################################
#
# Optional keywords
#
################################################################################
    cut_atomic_basis           .true.
    plus_u                     2 p 0.5
    plus_u                     3 d 0.25
    cut_core                   infinte 2.0
    innermost_max              5
################################################################################
#
#  Definition of "minimal" basis
#
################################################################################
#     valence basis states
    valence      3  s   2.
    valence      3  p   2.
#     ion occupancy
    ion_occ      3  s   1.
    ion_occ      3  p   1.
################################################################################
#
#  Suggested additional basis functions. For production calculations,
#  uncomment them one after another (the most important basis functions are
#  listed first).
#
#  Constructed for dimers: 1.75 A, 2.0 A, 2.25 A, 2.75 A, 3.75 A
#
################################################################################
#  "First tier" - improvements: -571.96 meV to -37.03 meV
     hydro 3 d 4.2
     hydro 2 p 1.4
     hydro 4 f 6.2
     ionic 3 s auto
#  "Second tier" - improvements: -16.76 meV to -3.03 meV
     hydro 3 d 9
     hydro 5 g 9.4
#     hydro 4 p 4
#     hydro 1 s 0.65
#  "Third tier" - improvements: -3.89 meV to -0.60 meV
#     ionic 3 d auto
#     hydro 3 s 2.6
#     hydro 4 f 8.4
#     hydro 3 d 3.4
#     hydro 3 p 7.8
#  "Fourth tier" - improvements: -0.33 meV to -0.11 meV
#     hydro 2 p 1.6
#     hydro 5 g 10.8
#     hydro 5 f 11.2
#     hydro 3 d 1
#     hydro 4 s 4.5
#  Further basis functions that fell out of the optimization - noise
#  level... < -0.08 meV
#     hydro 4 d 6.6
#     hydro 5 g 16.4
#     hydro 4 d 9
################################################################################
#
# For methods that use the localized form of the "resolution of identity" for
# the two-electron Coulomb operator (RI_method LVL), particularly Hartree-Fock and
# hybrid density functional calculations, the highest accuracy can be obtained by
# uncommenting the line beginning with "for_aux"  below, thus adding an extra g radial
# function to the construction of the product basis set for the expansion.
# See Ref. New J. Phys. 17, 093020 (2015) for more information, particularly Figs. 1 and 6.
#
################################################################################
#
# for_aux hydro 5 g 6.0""" # noqa: E501

species_out_content = """################################################################################
#
#  FHI-aims code project
#  Volker Blum, Fritz Haber Institute Berlin, 2009
#
#  Suggested "tight" defaults for Si atom (to be pasted into control.in file)
#
#  Revised Jan 04, 2011, following tests (SiC) done by Lydia Nemec:
#     d and g functions of tier 2 now enabled by default.
#
################################################################################

  species               Si
#     global species definitions
  nucleus               14.0
  mass                  28.0855
#
  l_hartree             6
#
  cut_pot               4.0 2.0 1.0
  basis_dep_cutoff      0.001
#
  radial_base    42   7.0000
  radial_multiplier        2
  angular_grids  specified
  division       0.4121   50
  division       0.7665  110
  division       1.0603  194
  division       1.2846  302
  division       1.4125  434
  outer_grid     434
##################################################################################
#
# Optional basis settings passed to pyfhiaims
#
##################################################################################
  cut_atomic_basis      True
  cut_core              infinte 2.0
  innermost_max         5
    plus_u    2 p 0.5
    plus_u    3 d 0.25
##################################################################################
#
#  Definition of "minimal" basis
#
##################################################################################
#     valence basis states
    valence      3  s   2.
    valence      3  p   2.
#     ion occupancy
    ion_occ      3  s   1.
    ion_occ      3  p   1.
##################################################################################
#
#  Suggested additional basis functions. For production calculations,
#  uncomment them one after another (the most important basis functions are
#  listed first).
#
#  These were set using pyfhiaims, original may files have additional comments
#
##################################################################################
#  First tier
             hydro      3  d  4.2
             hydro      2  p  1.4
             hydro      4  f  6.2
             ionic      3  s  auto
#  Second tier
             hydro      3  d  9.0
             hydro      5  g  9.4
#            hydro      4  p  4.0
#            hydro      1  s  0.65
#  Third tier
#            ionic      3  d  auto
#            hydro      3  s  2.6
#            hydro      4  f  8.4
#            hydro      3  d  3.4
#            hydro      3  p  7.8
#  Fourth tier
#            hydro      2  p  1.6
#            hydro      5  g  10.8
#            hydro      5  f  11.2
#            hydro      3  d  1.0
#            hydro      4  s  4.5
#  Further basis functions
#            hydro      4  d  6.6
#            hydro      5  g  16.4
#            hydro      4  d  9.0""" # noqa: E501


def test_species_defaults():
    """Test the species defaults"""
    speceis_default = SpeciesDefaults.from_strings(species_content_in.splitlines())

    assert str(speceis_default) == speceis_default.content
    assert speceis_default.content == species_out_content



"""An object representing FHI-aims input.

(control.in, geometry.in, k_list.in, T_bvk.in?)
"""

from dataclasses import dataclass
from pathlib import Path

from pyfhiaims.control import AimsControl
from pyfhiaims.control.kpoints import AimsKPoints
from pyfhiaims.errors import PyaimsError
from pyfhiaims.geometry import AimsGeometry
from pyfhiaims.species_defaults.species import SpeciesDefaults


@dataclass
class AimsInput:
    """An object combining all input files for FHI-aims and providing methods
    to deal with them consistently.
    """

    control: AimsControl
    geometry: AimsGeometry = None
    species_defaults: SpeciesDefaults | dict[str, SpeciesDefaults] = None

    @classmethod
    def from_folder(cls, folder: Path | str) -> "AimsInput":
        """Get AimsInput object from a given folder."""
        if isinstance(folder, str):
            folder = Path(folder)
        if not folder.is_dir():
            raise PyaimsError("The given path is not a directory.")

        control, geometry, k_points, species_defaults = [
            None,
        ] * 4
        if (folder / "control.in").is_file():
            control = AimsControl.from_file(folder / "control.in")
        else:
            raise PyaimsError(f"No control.in file found in {folder.as_posix()}.")

        return cls(
            control,
            geometry,
        )

    @property
    def k_points(self) -> AimsKPoints:
        """The k-points for FHI-aims calculations."""
        return self.control.k_points

    @k_points.setter
    def k_points(self, k_points: AimsKPoints) -> None:
        """Setter for the k-points of an FHI-aims calculation."""
        self.control.k_points = k_points



"""FHI-aims objects representing K-points (in the form of grid or list)."""

from dataclasses import dataclass
from enum import Enum


class KPointsType(Enum):
    """Enum type for Kpoints."""

    GRID = "k_grid"
    DENSITY = "k_grid_density"
    EXTERNAL = "k_points_external"


@dataclass
class AimsKPoints:
    """A base class for FHI-aims K-points representation."""

    type: KPointsType



"""Utility functions for pyaims test suite."""

import yaml

from pyfhiaims.outputs.parser import StdoutParser


def prepare_ref_data(data_dir):
    """Prepare reference data for the output files in `data_dir`.

    Args:
        data_dir : str

    """

    def remove_key(data, key_to_remove):
        if isinstance(data, dict):
            # Remove the key if it exists, then recursively check values
            data.pop(key_to_remove, None)
            for value in data.values():
                remove_key(value, key_to_remove)
        elif isinstance(data, list):
            # Apply the function to each item in the list
            for item in data:
                remove_key(item, key_to_remove)

    for file_name in data_dir.iterdir():
        if file_name.is_file():
            parser = StdoutParser(file_name)
            results = parser.parse()
            remove_key(results, "geometry")
            remove_key(results, "start_time")
            remove_key(results, "end_time")
            remove_key(results, "input")
            remove_key(results, "self_energy")

            with open(data_dir / "ref" / (file_name.stem + ".yaml"), "w") as outfile:
                yaml.dump(results, outfile, default_flow_style=False)


def is_subset(json1, json2):
    """Check if json1 is a subset of a (larger) json2."""

    def _is_subset_list(list1, list2):
        """Check if all dicts in list1 are subsets of dicts in list2."""
        if len(list1) != len(list2):
            return False
        for el1, el2 in zip(list1, list2, strict=False):
            if type(el1) is not type(el2):
                return False
            if isinstance(el1, dict):
                if not _is_subset_dict(el1, el2):
                    return False
            elif isinstance(el1, list):
                if not _is_subset_list(el1, el2):
                    return False
            elif el1 != el2:
                return False
        return True

    def _is_subset_dict(dict1, dict2):
        """Check if dict1 is a subset of dict2."""
        for k, v in dict1.items():
            if k not in dict2:
                return False
            if type(v) is not type(dict2[k]):
                return False
            if isinstance(v, dict):
                if not _is_subset_dict(v, dict2[k]):
                    return False
            elif isinstance(v, list):
                if not _is_subset_list(v, dict2[k]):
                    return False
            elif v != dict2[k]:
                return False
        return True

    return _is_subset_dict(json1, json2)



"""Type casting utils from pymatgen that are useful."""

from collections.abc import Sequence
from typing import Annotated, Literal, TypeVar

import numpy as np
import numpy.typing as npt

_T = TypeVar("_T", bound=np.generic)

Vector3D = Annotated[npt.NDArray[_T], Literal[3]]
Matrix3D = Annotated[npt.NDArray[_T], Literal[3]]


def to_vector3d(sequence: Sequence, dtype=float) -> Vector3D:
    """Type casting to 3D vector of floats."""
    if len(sequence) != 3:
        raise ValueError("Vector3D requires a length of 3")
    return np.array(sequence, dtype=dtype)


def to_matrix3d(sequence: Sequence, dtype=float) -> Matrix3D:
    """Type casting to 3D matrix of floats."""
    if (len(sequence) != 3) and all(len(row) != 3 for row in sequence):
        raise ValueError("Matrix3D requires a length of 3 in both dimsensions.")

    return np.array(sequence, dtype=dtype)



"""Tests for IntegrationGrid class."""

from textwrap import dedent

from pyfhiaims.species_defaults.integration_grid import IntegrationGrid


def test_integration_grid():
    """Test integration grids."""
    test_string = dedent(
        """\
    radial_base         24 7.0
    radial_multiplier   8
    angular_grids       specified
    division   0.1930   50
    division   0.3175  110
    division   0.4293  194
    division   0.5066  302
    division   0.5626  434
    division   0.5922  590
    #    division   0.6227  974
    #    division   0.6868 1202
    outer_grid  770
    #    outer_grid  434
    """
    )
    grid = IntegrationGrid.from_string(test_string)
    assert grid.radial_grid.radius == 7.0
    assert len(grid.angular_grid.r) == 6



"""Input sets for band structure calculations."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, TypedDict

from pymatgen.core import Molecule, Structure
from pymatgen.io.aims.sets.base import AimsInputGenerator
from pymatgen.symmetry.bandstructure import HighSymmKpath

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any


class _SegmentDict(TypedDict):
    coords: list[list[float]]
    labels: list[str]
    length: int


def prepare_band_input(structure: Structure, density: float = 20):
    """Prepare the band information needed for the FHI-aims control.in file.

    Args:
        structure (Structure): The structure for which the band path is calculated
        density (float): Number of kpoints per Angstrom.
    """
    bp = HighSymmKpath(structure)
    points, labels = bp.get_kpoints(line_density=density, coords_are_cartesian=False)
    lines_and_labels: list[_SegmentDict] = []
    current_segment: _SegmentDict | None = None
    for label_, coords in zip(labels, points, strict=True):
        # rename the Gamma point label
        label = "G" if label_ in ("GAMMA", "\\Gamma", "Γ") else label_
        if label:
            if current_segment is None:
                current_segment = _SegmentDict(coords=[coords], labels=[label], length=1)
            else:
                current_segment["coords"].append(coords)
                current_segment["labels"].append(label)
                current_segment["length"] += 1
                lines_and_labels.append(current_segment)
                current_segment = None
        elif current_segment is not None:
            current_segment["length"] += 1

    bands = []
    for segment in lines_and_labels:
        start, end = segment["coords"]
        label_start, label_end = segment["labels"]
        bands.append(
            f"band {start[0]:9.5f}{start[1]:9.5f}{start[2]:9.5f} "
            f"{end[0]:9.5f}{end[1]:9.5f}{end[2]:9.5f} {segment['length']:4d} "
            f"{label_start:3}{label_end:3}"
        )
    return bands


@dataclass
class BandStructureSetGenerator(AimsInputGenerator):
    """A generator for the band structure calculation input set.

    Attributes:
        calc_type (str): The type of calculations
        k_point_density (float): The number of k_points per angstrom
    """

    calc_type: str = "bands"
    k_point_density: float = 20

    def get_parameter_updates(
        self, structure: Structure | Molecule, prev_parameters: dict[str, Any]
    ) -> dict[str, Sequence[str]]:
        """Get the parameter updates for the calculation.

        Args:
            structure (Structure): The structure to calculate the bands for
            prev_parameters (Dict[str, Any]): The previous parameters

        Returns:
            dict: The updated for the parameters for the output section of FHI-aims
        """
        if isinstance(structure, Molecule):
            raise TypeError("BandStructures can not be made for non-periodic systems")

        updated_outputs = prev_parameters.get("output", [])
        updated_outputs += prepare_band_input(structure, self.k_point_density)
        return {"output": updated_outputs}


@dataclass
class GWSetGenerator(AimsInputGenerator):
    """
    A generator for the input set for calculations employing GW self-energy correction.

    Attributes:
        calc_type (str): The type of calculations
        k_point_density (float): The number of k_points per angstrom
    """

    calc_type: str = "GW"
    k_point_density: float = 20

    def get_parameter_updates(self, structure: Structure | Molecule, prev_parameters: dict[str, Any]) -> dict[str, Any]:
        """Get the parameter updates for the calculation.

        Args:
            structure (Structure or Molecule): The structure to calculate the bands for
            prev_parameters (Dict[str, Any]): The previous parameters

        Returns:
            dict: The updated for the parameters for the output section of FHI-aims
        """
        updates = {"anacon_type": "two-pole"}
        current_output = prev_parameters.get("output", [])
        if isinstance(structure, Structure) and all(structure.lattice.pbc):
            updates.update(
                qpe_calc="gw_expt",
                output=current_output + prepare_band_input(structure, self.k_point_density),
            )
        else:
            updates.update(qpe_calc="gw")
        return updates



"""Test the AimsCube interface"""

import numpy as np
import pytest

from pyfhiaims.geometry.atom import FHIAimsAtom

content_base = """atom         0.000000000000e+00   1.100000000000e+00  -2.500000000000e-01 H
    velocity   1.000000000000e+00   2.000000000000e+00   3.000000000000e+00
    initial_charge 0.200000000000
    initial_moment 0.500000000000
    constrain_relaxation x
    constraint_region 1
    magnetic_moment 0.200000000000
    nuclear_spin 0.300000000000
    isotope 3
    RT_TDDFT_initial_velocity   1.000000000000e+00   2.000000000000e+00   3.000000000000e+00"""  # noqa: E501

content_frac = """atom_frac    0.000000000000e+00   1.100000000000e+00  -2.500000000000e-01 H
    velocity   1.000000000000e+00   2.000000000000e+00   3.000000000000e+00
    initial_charge 0.200000000000
    initial_moment 0.500000000000
    constrain_relaxation x
    constraint_region 1
    magnetic_moment 0.200000000000
    nuclear_spin 0.300000000000
    isotope 3
    RT_TDDFT_initial_velocity   1.000000000000e+00   2.000000000000e+00   3.000000000000e+00"""  # noqa: E501

content_frac_wrap = """atom_frac    0.000000000000e+00   1.000000000000e-01   7.500000000000e-01 H
    velocity   1.000000000000e+00   2.000000000000e+00   3.000000000000e+00
    initial_charge 0.200000000000
    initial_moment 0.500000000000
    constrain_relaxation x
    constraint_region 1
    magnetic_moment 0.200000000000
    nuclear_spin 0.300000000000
    isotope 3
    RT_TDDFT_initial_velocity   1.000000000000e+00   2.000000000000e+00   3.000000000000e+00"""  # noqa: E501

content_empty = """empty        0.000000000000e+00   1.100000000000e+00  -2.500000000000e-01 H
    velocity   1.000000000000e+00   2.000000000000e+00   3.000000000000e+00
    initial_charge 0.200000000000
    initial_moment 0.500000000000
    constrain_relaxation x
    constraint_region 1
    magnetic_moment 0.200000000000
    nuclear_spin 0.300000000000
    isotope 3
    RT_TDDFT_initial_velocity   1.000000000000e+00   2.000000000000e+00   3.000000000000e+00"""  # noqa: E501

content_psuedocore = """pseudocore   0.000000000000e+00   1.100000000000e+00  -2.500000000000e-01 H
    velocity   1.000000000000e+00   2.000000000000e+00   3.000000000000e+00
    initial_charge 0.200000000000
    initial_moment 0.500000000000
    constrain_relaxation x
    constraint_region 1
    magnetic_moment 0.200000000000
    nuclear_spin 0.300000000000
    isotope 3
    RT_TDDFT_initial_velocity   1.000000000000e+00   2.000000000000e+00   3.000000000000e+00"""  # noqa: E501


def test_atom(tmp_path):
    """Test the AimsCube Interface"""
    atom = FHIAimsAtom(
        symbol="H",
        position=[0.0, 1.1, -0.25],
        velocity=[1, 2, 3],
        initial_charge=0.2,
        initial_moment=0.5,
        constraints=(True, False, False),
        constraint_region=1,
        magnetic_response=False,
        magnetic_moment=0.2,
        nuclear_spin=0.3,
        isotope=3,
        is_empty=False,
        is_pseudocore=False,
        RT_TDDFT_initial_velocity=[1, 2, 3],
    )

    assert atom.to_string() == content_base

    atom.set_fractional(np.eye(3))
    assert atom.to_string() == content_frac

    atom.set_fractional(np.eye(3), wrap=True)
    assert atom.to_string() == content_frac_wrap

    atom.is_empty = True
    assert atom.to_string() == content_empty

    atom.is_empty = False
    atom.is_pseudocore = True
    assert atom.to_string() == content_psuedocore

    with pytest.raises(KeyError):
        _ = FHIAimsAtom(symbol="Asdf", position=[0, 0, 0])



"""Various utilities for pyaims."""


def is_number_string(s: str) -> bool:
    """Check if s is a string made of float numbers and optionally comment sign.

    Parameters
    ----------
    s: str
        A string to check

    Returns
    -------
    bool
        True if the string represents a number

    """
    allowed_chars = set("-.# ")
    return all(char.isdigit() or char in allowed_chars for char in s)



"""A dataclass representing integration grid in species' defaults file."""

from collections.abc import Sequence
from dataclasses import dataclass, field

__all__ = [
    "IntegrationGrid",
]

from pyfhiaims.control.chunk import AimsControlChunk


@dataclass
class RadialIntegrationGrid(AimsControlChunk):
    r"""Radial part of integration grid.

    Parameters
    ----------
    number: int
        The total number of grid points N.
    radius: float
        The outermost shell radius $r_{outer}$ of the basic grid in \\AA.
        The location of the i-th radial shell is given by
        \\[ r_{i} = r_{outer} \\cdot \frac{1-[i/(N+1)]^2}{1-[N/(N+1)]^2} ]\
    multiplier: int
        The number of added subdivisions per basic grid spacing

    """

    number: int = field(default_factory=int)
    radius: float = field(default_factory=float)
    multiplier: int = 2
    keywords: tuple[str] = ("radial_base", "radial_multiplier")

    @classmethod
    def from_strings(cls, strings: Sequence[str]) -> "RadialIntegrationGrid":
        """Get a RadialIntegrationGrid instance from a sequence of strings."""
        num = 0
        r = 0.0
        multiplier = 2
        for line in strings:
            line = line.strip()
            if "radial_base" in line:
                num = int(line.split()[1])
                r = float(line.split()[2])
            if "radial_multiplier" in line:
                multiplier = int(line.split()[1])
        return cls(number=num, radius=r, multiplier=multiplier)

    @classmethod
    def from_string(cls, string: str) -> "RadialIntegrationGrid":
        return cls.from_strings(string.splitlines())

    def to_string(self) -> str:
        """Gt the string representation of the AngularIntegrationGrid."""
        result = [
            f"  radial_base  {self.number:4d}   {self.radius:6.4f}",
            f"  radial_multiplier   {self.multiplier:6d}",
        ]
        return "\n".join(result)


@dataclass
class AngularIntegrationGrid(AimsControlChunk):
    """Angular part of integration grid.

    Args:
    # angular: Optional[int]
    #     For self-adapting angular integration grids, the maximum allowed number of
    #     points per radial shell.
    # angular_acc: Optional[float]
    #     For self-adapting angular integration grids, specifies the desired
    #     integration accuracy for the initial Hamiltonian and overlap matrix elements
    #     Use only for cluster-type geometries, if 0 no adaptation is performed
    # angular_min: Optional[int]
    #     Specifies the minimum number of angular grid points per radial
    #     integration shell
    type: str
        Indicates how the angular integration grids (in each radial integration
        shell) for this species are determined. method is a string,
        either auto or specified.
    n_outer: int
        Specify the number of angular grid points used outside the outermost
        division radius

    """

    type: str = "specified"
    r: Sequence[float] = field(default_factory=list)
    n: Sequence[int] = field(default_factory=list)
    n_outer: int = field(default_factory=int)
    keywords: tuple[str] = (
        "angular_grids",
        "division",
        "outer_grid",
        "angular",
        "angular_acc",
        "angular_min",
    )

    def __post_init__(self):
        """Perform sanity checks on an object."""
        if len(self.r) != len(self.n):
            raise ValueError("r and n grids should be the same")

    @classmethod
    def from_strings(cls, strings: Sequence[str]) -> "AngularIntegrationGrid":
        """Get an AngularIntegrationGrid from a list of strings."""
        cls_type = "specified"
        r = []
        n = []
        n_outer = 0
        for line in strings:
            line = line.strip()
            if line.startswith("#"):
                continue
            if "angular_grids" in line:
                cls_type = line.split()[1]
            if "division" in line:
                r.append(float(line.split()[1]))
                n.append(int(line.split()[2]))
            if "outer_grid" in line:
                n_outer = int(line.split()[1])
        return cls(type=cls_type, r=r, n=n, n_outer=n_outer)

    @classmethod
    def from_string(cls, string: str) -> "AngularIntegrationGrid":
        """Get an AngularIntegrationGrid from a part of species' defaults file."""
        return cls.from_strings(string.splitlines())

    def to_string(self) -> str:
        """Get the string representation of the AngularIntegrationGrid."""
        result = [
            f"  angular_grids  {self.type}",
        ]
        for r_i, n_i in zip(self.r, self.n, strict=False):
            result.append(f"  division       {r_i:6.4f} {n_i:4d}".rstrip())
        result.append(f"  outer_grid     {self.n_outer:<4d}".rstrip())
        return "\n".join(result)


@dataclass
class IntegrationGrid(AimsControlChunk):
    """The integration grid class."""

    radial_grid: RadialIntegrationGrid
    angular_grid: AngularIntegrationGrid

    @classmethod
    def from_strings(cls, strings: Sequence[str]) -> "IntegrationGrid":
        """Get the integration grid from a list of strings."""
        radial_str = [
            s for s in strings if any(k in s for k in RadialIntegrationGrid.keywords)
        ]
        angular_str = [
            s for s in strings if any(k in s for k in AngularIntegrationGrid.keywords)
        ]
        return cls(
            radial_grid=RadialIntegrationGrid.from_strings(radial_str),
            angular_grid=AngularIntegrationGrid.from_strings(angular_str),
        )

    @classmethod
    def from_string(cls, string: str) -> "IntegrationGrid":
        """Get the integration grid from a string."""
        return cls.from_strings(string.splitlines())

    def to_string(self) -> str:
        """Get the string representation of the integration grid."""
        return "\n".join([str(self.radial_grid), str(self.angular_grid)])



"""A simple parser written to test the outputs.parser approach."""

from pyfhiaims.outputs.parser import converters
from pyfhiaims.outputs.parser.abc import EXP, FLOAT, INT, ChunkParser, FileParser


class SimpleParser(FileParser):
    """A simple parser written to test the outputs.parser approach."""
    name = "root"
    initial_chunk = "a"


class AParser(ChunkParser):
    name = "a"
    values = {
        "version": r"Version: (\S*)"
    }
    metadata = {}
    next_chunks = ["b", "c"]
    parsed_key = ""


class BParser(ChunkParser):
    name = "b"
    title_line = "B CHUNK"
    values = {
        "b1": converters.to_int(fr" \| B1 = ({INT})"),
        "b2": converters.to_float(fr" \| B2 = {FLOAT} Ry ({FLOAT}) eV"),
        "b3": converters.to_float(fr"B3 = ({EXP})"),
        "b4": converters.to_float(fr" \| B4 = ({INT})")
    }
    next_chunks = ["c"]
    parsed_key = "b_values"


class CParser(ChunkParser):
    name = "c"
    title_line = "--- C "
    values = {
        "t": converters.to_float(fr"time = ({FLOAT})"),
        "all_good": converters.to_bool(r"Everything is good"),
    }
    metadata = {}
    next_chunks = ["c", "d"]
    parsed_key = "b_values.c_values[]"


class DParser(ChunkParser):
    name = "d"
    title_line = "Converged"
    values = {
        "end_time": converters.to_time(rf"End time: ({FLOAT})")
    }
    metadata = {}
    parsed_key = "b_values.c_values"



"""Some tests for the AimsStdout object."""

from datetime import date
from math import isclose

import pytest

from pyfhiaims import AimsStdout
from pyfhiaims.errors import AimsParseError
from pyfhiaims.outputs.parser import StdoutParser, converters


def test_stdout_empty(data_dir):
    file_name = data_dir / "stdout" / "output_files" / "preamble_fail.out"
    with pytest.raises(AimsParseError, match="is not a valid FHI-aims output file"):
        _ = AimsStdout(file_name)


@pytest.mark.parametrize("file_name", ["no_control_in.out", "low_stack.out"])
def test_stdout_no_start_warnings(data_dir, file_name):
    """Test raising warnings with incomplete control.in."""
    file_path = data_dir / "stdout" / "output_files" / file_name

    with pytest.warns(UserWarning, match="FHI-aims calculation did not start"):
        _ = AimsStdout(file_path)

    # with pytest.warns(UserWarning, match="No SCF steps found"):
    #     _ = AimsStdout(file_path)


def test_stdout_no_control_in(data_dir):
    """Test parsing output without input files."""
    file_name = data_dir / "stdout" / "output_files" / "no_control_in.out"
    stdout = AimsStdout(file_name)
    assert stdout.aims_version == "240910"
    assert stdout.start_date == date(2025, 1, 10)
    assert not stdout.is_finished_ok
    assert len(stdout.errors) == 1
    assert "Input file control.in not found" in stdout.errors[0]
    # no final values
    assert stdout.energy is None
    assert all(v is None for _, v in stdout.results["final"].items())
    # we got to input, but found nothing
    assert not stdout.results["input"]


def test_stdout_low_stack(data_dir):
    """Test parsing output with low stack size."""
    file_name = data_dir / "stdout" / "output_files" / "low_stack.out"
    stdout = AimsStdout(file_name)
    assert stdout.aims_version == "240910"
    assert not stdout.is_finished_ok
    assert len(stdout.errors) == 1
    assert "Current stacksize too low" in stdout.errors[0]
    assert len(stdout.warnings) == 2
    assert stdout.energy is None
    assert all(v is None for _, v in stdout.results["final"].items())
    assert stdout.metadata.k_grid == "10 10 10"


def test_stdout(data_dir):
    """Test stdout."""
    stdout_data_dir = data_dir / "stdout"
    file_name = stdout_data_dir / "relax.out.gz"
    stdout = AimsStdout(file_name)
    assert len(stdout.warnings) == 1
    assert len(stdout.errors) == 0
    assert stdout.is_finished_ok
    assert stdout.geometry_converged
    assert isclose(stdout.energy, -1071304.321815)
    assert stdout.forces is not None
    assert stdout.metadata.num_atoms == 2


def test_stdout_own_values(data_dir):
    """Test stdout with own values."""
    stdout_data_dir = data_dir / "stdout"
    file_name = stdout_data_dir / "relax.out.gz"
    parser = StdoutParser(file_name)
    parser.add_parsed_values(
        "scf_step",
        cpu_time=converters.to_float(r"Time for this iteration *: *([.\d]+) s"),
    )
    stdout = AimsStdout(file_name, parser=parser)
    for ionic_step in stdout.results["ionic_steps"]:
        for scf_step in ionic_step["scf_steps"][:-1]:
            assert "cpu_time" in scf_step
    assert stdout.results["ionic_steps"][0]["scf_steps"][0]["cpu_time"] == 0.165


def test_stdout_scf_analysis(data_dir):
    """Test stdout with own values."""
    file_name = data_dir / "stdout" / "output_files" / "scf_analysis.out"
    stdout = AimsStdout(file_name)
    assert stdout.aims_version == "250626"
    assert stdout.is_finished_ok
    assert stdout.num_scf_steps == 53

    for i, scf_step in enumerate(stdout.results["ionic_steps"][0]["scf_steps"]):
        if "charge_density_status" in scf_step:
            assert i in [24, 49]
            if i == 24:
                assert scf_step["estimated_steps_to_convergence"] == 28
                assert scf_step["convergence_expected"]
                assert (
                    scf_step["charge_density_status"]
                    == "Good convergence trend detected"
                )
            if i == 49:
                assert scf_step["estimated_steps_to_convergence"] == 3
                assert scf_step["convergence_expected"]
                assert (
                    scf_step["charge_density_status"]
                    == "Good convergence trend detected"
                )


def test_stdout_scf_analysis_spin(data_dir):
    """Test stdout with own values."""
    file_name = data_dir / "stdout" / "output_files" / "scf_analysis_spin.out"
    stdout = AimsStdout(file_name)
    assert stdout.aims_version == "250626"
    assert stdout.is_finished_ok
    assert stdout.num_scf_steps == 53

    for i, scf_step in enumerate(stdout.results["ionic_steps"][0]["scf_steps"]):
        if "charge_density_status" and "spin_density_status" in scf_step:
            assert i in [24, 49]
            if i == 24:
                assert scf_step["estimated_steps_to_convergence"] == 28
                assert scf_step["convergence_expected"]
                assert (
                    scf_step["charge_density_status"]
                    == "Good convergence trend detected"
                )
                assert scf_step["spin_density_status"] == "Already converged"
            if i == 49:
                assert scf_step["estimated_steps_to_convergence"] == 3
                assert scf_step["convergence_expected"]
                assert (
                    scf_step["charge_density_status"]
                    == "Good convergence trend detected"
                )
                assert scf_step["spin_density_status"] == "Already converged"



"""A class representing BandStructure."""

from dataclasses import dataclass

import numpy as np


@dataclass
class BandStructure:
    """A band structure."""

    is_soc: bool
    is_mulliken: bool
    data: np.ndarray



"""A representation of FHI-aims output (based on ASE output parser)."""

from __future__ import annotations

import warnings
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from monty.json import MontyDecoder, MSONable
from pymatgen.core import Lattice, Structure

from pyfhiaims.outputs.stdout import AimsParseError, AimsStdout

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any, Self

    from pymatgen.core import Molecule
    from pymatgen.util.typing import Matrix3D, Vector3D

__author__ = "Andrey Sobolev and Thomas A. R. Purcell"
__version__ = "1.0"
__email__ = "andrey.n.sobolev@gmail.com and purcellt@arizona.edu"
__date__ = "November 2023"


AIMS_OUTPUT_KEY_MAP = {
    "free_energy": "energy",  # TARP These are the force consistent energies
}


def remap_outputs(results: dict[str, Any]) -> dict[str, Any]:
    """Remap FHIAimsOutput keys to AimsOutput keys"""
    to_ret = results.copy()
    for key, val in AIMS_OUTPUT_KEY_MAP.items():
        to_ret[val] = to_ret.pop(key, None)

    return to_ret


class AimsOutput(MSONable):
    """The main output file for FHI-aims."""

    def __init__(
        self,
        structures: Molecule | Structure | Sequence[Molecule | Structure],
        results: dict[str, Any],
        metadata: dict[str, Any],
        structure_summary: dict[str, Any],
        warnings: list[str] | None = None,
        errors: list[str] | None = None,
    ) -> None:
        """
        Args:
            structures (Molecule or Structure or Sequence[Molecule or Structure]):  A list
                of all images in an output file
            results (Dict[str, Any]): A dictionary of all parsed results from
                the output file.
            metadata (Dict[str, Any]): The metadata of the executable used to perform
                the calculation
            structure_summary (Dict[str, Any]): The summary of the starting
                atomic structure.
            warnings (List[str]): A list of warnings from the calculation
            errors (List[str]): A list of errors from the calculation
        """
        self._structures = structures
        self._results = results
        self._metadata = metadata
        self._structure_summary = structure_summary
        self._warnings = warnings
        self._errors = errors

    def as_dict(self) -> dict[str, Any]:
        """Create a dict representation of the outputs for MSONable."""
        dct: dict[str, Any] = {
            "@module": type(self).__module__,
            "@class": type(self).__name__,
            "structures": self._structures,
            "metadata": self._metadata,
            "structure_summary": self._structure_summary,
            "warnings": self._warnings,
            "errors": self._errors,
        }
        return dct

    @classmethod
    def from_outfile(cls, outfile: str | Path) -> Self:
        """Construct an AimsOutput from an output file.

        Args:
            outfile: str | Path: The aims.out file to parse

        Returns:
            The AimsOutput object for the output file

        Raises:
            AimsParseError if a file does not exist
        """
        aims_out = None
        for path in [Path(outfile), Path(f"{outfile}.gz")]:
            try:
                aims_out = AimsStdout(path)
            except FileNotFoundError:
                continue

        if aims_out is None:
            raise AimsParseError(f"File {outfile} not found.")

        # remove all the numbers from metadata and put them into structure (more like input) summary
        metadata = {k: v for k, v in aims_out.metadata.items() if not k.startswith("num_")}

        structure_summary = aims_out.header_summary
        structure_summary.update(
            **{k.replace("num_", "n_"): v for k, v in aims_out.metadata.items() if k.startswith("num_")}
        )

        structure_summary["initial_structure"] = structure_summary.pop("initial_geometry").structure
        for site in structure_summary["initial_structure"]:
            if abs(site.properties.get("magmom", 0.0)) < 1e-10:
                site.properties.pop("magmom", None)

        lattice = structure_summary.pop("initial_lattice", None)
        if lattice is not None:
            lattice = Lattice(lattice)
        structure_summary["initial_lattice"] = lattice

        results = []
        for image in aims_out:
            image_results = remap_outputs(image.results)
            site_prop_keys = {
                "forces": "force",
                "stresses": "atomic_virial_stress",
                "hirshfeld_charges": "hirshfeld_charge",
                "hirshfeld_volumes": "hirshfeld_volume",
                "hirshfeld_atomic_dipoles": "hirshfeld_atomic_dipole",
                "mulliken_charges": "charge",
                "mulliken_spins": "magmom",
            }
            properties = {prop: image_results[prop] for prop in image_results if prop not in site_prop_keys}
            site_properties = {}
            for prop, site_key in site_prop_keys.items():
                if prop in image_results:
                    site_properties[site_key] = image_results[prop]

            if ((magmom := site_properties.get("magmom")) is not None) and np.abs(
                np.sum(magmom) - properties["magmom"]
            ) < 1e-3:
                warnings.warn(
                    "Total magnetic moment and sum of Mulliken spins are not consistent",
                    stacklevel=2,
                )
            structure = image.geometry.to_structure(properties=properties, site_properties=site_properties)
            results.append(structure)

        return cls(results, aims_out.results, metadata, structure_summary, aims_out.warnings, aims_out.errors)

    @classmethod
    def from_dict(cls, dct: dict[str, Any]) -> Self:
        """Construct an AimsOutput from a dictionary.

        Args:
            dct (dict[str, Any]): The dictionary used to create AimsOutput

        Returns:
            AimsOutput
        """
        decoded = {k: MontyDecoder().process_decoded(v) for k, v in dct.items() if not k.startswith("@")}
        for struct in decoded["structures"]:
            struct.properties = {k: MontyDecoder().process_decoded(v) for k, v in struct.properties.items()}

        return cls(
            decoded["structures"],
            decoded["results"],
            decoded["metadata"],
            decoded["structure_summary"],
            decoded["warnings"],
            decoded["errors"],
        )

    def get_results_for_image(self, image_ind: int) -> Structure | Molecule:
        """Get the results dictionary for a particular image or slice of images.

        Args:
            image_ind (int): The index of the image to get the results for

        Returns:
            The results of the image with index images_ind
        """
        return self._structures[image_ind]

    @property
    def structure_summary(self) -> dict[str, Any]:
        """The summary of the material/molecule that the calculations represent."""
        return self._structure_summary

    @property
    def metadata(self) -> dict[str, Any]:
        """The system metadata."""
        return self._metadata

    @property
    def n_images(self) -> int:
        """The number of images in results."""
        return len(self._structures)

    @property
    def initial_structure(self) -> Structure | Molecule:
        """The initial structure for the calculations."""
        return self._structure_summary["initial_structure"]

    @property
    def final_structure(self) -> Structure | Molecule:
        """The final structure for the calculation."""
        return self._structures[-1]

    @property
    def structures(self) -> Sequence[Structure | Molecule]:
        """All images in the output file."""
        return self._structures

    @property
    def results(self) -> dict[str, Any]:
        """All results for the calculation."""
        return self._results

    @property
    def fermi_energy(self) -> float:
        """The Fermi energy for the final structure in the calculation."""
        return self.get_results_for_image(-1).properties["fermi_energy"]

    @property
    def vbm(self) -> float:
        """The HOMO level for the final structure in the calculation."""
        return self.get_results_for_image(-1).properties["vbm"]

    @property
    def cbm(self) -> float:
        """The LUMO level for the final structure in the calculation."""
        return self.get_results_for_image(-1).properties["cbm"]

    @property
    def band_gap(self) -> float:
        """The band gap for the final structure in the calculation."""
        return self.get_results_for_image(-1).properties["gap"]

    @property
    def direct_band_gap(self) -> float:
        """The direct band gap for the final structure in the calculation."""
        return self.get_results_for_image(-1).properties["direct_gap"]

    @property
    def final_energy(self) -> float:
        """The total energy for the final structure in the calculation."""
        return self.get_results_for_image(-1).properties["energy"]

    @property
    def completed(self) -> bool:
        """Did the calculation complete."""
        return len(self._structures) > 0

    @property
    def aims_version(self) -> str:
        """The version of FHI-aims used for the calculation."""
        return self._metadata["aims_version"]

    @property
    def forces(self) -> Sequence[Vector3D] | None:
        """The forces for the final image of the calculation."""
        force_array = self.get_results_for_image(-1).site_properties.get("force", None)
        if isinstance(force_array, np.ndarray):
            return force_array.tolist()

        return force_array

    @property
    def stress(self) -> Matrix3D:
        """The stress for the final image of the calculation."""
        return self.get_results_for_image(-1).properties.get("stress", None)

    @property
    def stresses(self) -> Sequence[Matrix3D] | None:
        """The atomic virial stresses for the final image of the calculation."""
        stresses_array = self.get_results_for_image(-1).site_properties.get("atomic_virial_stress", None)
        if isinstance(stresses_array, np.ndarray):
            return stresses_array.tolist()
        return stresses_array

    @property
    def all_forces(self) -> list[list[Vector3D]]:
        """The forces for all images in the calculation."""
        all_forces_array = [res.site_properties.get("force", None) for res in self._structures]
        return [af.tolist() if isinstance(af, np.ndarray) else af for af in all_forces_array]



"""Module defining core FHI-aims input set generators."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from pymatgen.core import Structure
from pymatgen.io.aims.sets.base import AimsInputGenerator

if TYPE_CHECKING:
    from typing import Any

    from pymatgen.core import Molecule


_valid_dynamics: dict[str, tuple[str, ...]] = {
    "nve": ("", "4th_order", "damped"),
    "nvt": ("andersen", "berendsen", "parrinello", "nose-hoover"),
    "gle": ("thermostat",),
}


@dataclass
class StaticSetGenerator(AimsInputGenerator):
    """Common class for ground-state generators.

    Attributes:
        calc_type (str): The type of calculation
    """

    calc_type: str = "static"

    def get_parameter_updates(self, structure: Structure | Molecule, prev_parameters: dict[str, Any]) -> dict[str, Any]:
        """Get the parameter updates for the calculation.

        Args:
            structure (Structure or Molecule): The structure to calculate the bands for
            prev_parameters (Dict[str, Any]): The previous parameters

        Returns:
            dict: The updated for the parameters for the output section of FHI-aims
        """
        return prev_parameters


@dataclass
class RelaxSetGenerator(AimsInputGenerator):
    """Generate FHI-aims relax sets for optimizing internal coordinates and lattice params.

    Attributes:
        calc_type (str): The type of calculation
        relax_cell (bool): If True then relax the unit cell from the structure
        max_force (float): Maximum allowed force in the calculation
        method (str): Method used for the geometry optimization
    """

    calc_type: str = "relaxation"
    relax_cell: bool = True
    max_force: float = 1e-3
    method: str = "trm"

    def get_parameter_updates(self, structure: Structure | Molecule, prev_parameters: dict[str, Any]) -> dict:
        """Get the parameter updates for the calculation.

        Args:
            structure (Structure or Molecule): The structure to calculate the bands for
        prev_parameters (Dict[str, Any]): The previous parameters

        Returns:
            dict: The updated for the parameters for the output section of FHI-aims
        """
        updates = {"relax_geometry": f"{self.method} {self.max_force:e}"}
        if isinstance(structure, Structure) and self.relax_cell:
            updates["relax_unit_cell"] = "full"
        elif isinstance(structure, Structure):
            updates["relax_unit_cell"] = "none"

        prev_parameters.update(updates)
        return prev_parameters


@dataclass
class SocketIOSetGenerator(AimsInputGenerator):
    """Generate FHI-aims input sets for running with the socket.

    Attributes:
        calc_type (str): The type of calculation
        host (str): The hostname for the server the socket is on
        port (int): The port the socket server is listening on
    """

    calc_type: str = "multi_scf"
    host: str = "localhost"
    port: int = 12345

    def get_parameter_updates(self, structure: Structure | Molecule, prev_parameters: dict[str, Any]) -> dict:
        """Get the parameter updates for the calculation.

        Args:
            structure (Structure or Molecule): The structure to calculate the bands for
            prev_parameters (Dict[str, Any]): The previous parameters

        Returns:
            dict: The updated for the parameters for the output section of FHI-aims
        """
        return {"use_pimd_wrapper": (self.host, self.port)}


@dataclass
class MDSetGenerator(AimsInputGenerator):
    """
    A class for generating FHI-aims input sets for molecular dynamics calculations.

    Parameters
    ----------
    ensemble
        Molecular dynamics ensemble to run. Options include `nvt`, `nve`, and `gle`.
        Default: `nve`
    ensemble_specs
        A dictionary containing the specifications of the molecular dynamics ensemble.
        Valid keys are `type` (the ensemble type, valid types are defined in
        `_valid_dynamics` dict), and `parameter` - the control parameter for the thermostat
        (not used for `nve` and `nve_4th_order`).
    temp
        Thermostat temperature. Default: None
    time
        Simulation time (in picoseconds). Negative value stands for indefinite run.
        Default: 5 ps
    time_step
        The time step (in picoseconds) for the simulation. default: 1 fs
    **kwargs
        Other keyword arguments that will be passed to :obj:`AimsInputGenerator`.
    """

    calc_type: str = "md"
    ensemble: str = "nve"
    ensemble_specs: dict[str, Any] = field(default_factory=dict)
    temp: float | None = None
    time: float = 5.0
    time_step: float = 0.001
    init_velocities: bool = True

    def get_parameter_updates(self, structure: Structure | Molecule, prev_parameters: dict[str, Any]) -> dict:
        """Get the parameter updates for the calculation.

        Parameters
        ----------
        structure (Structure or Molecule):
            The structure to calculate the bands for
        prev_parameters (Dict[str, Any]):
            The previous parameters

        Returns
        -------
        dict
            A dictionary of updates to apply.
        """
        updates: dict[str, Any] = {
            "MD_run": [self.time],
            "MD_time_step": self.time_step,
        }

        # check for ensemble type validity
        default_ensemble_types = {"nve": "", "nvt": "parrinello", "gle": "thermostat"}
        if self.ensemble not in _valid_dynamics:
            raise ValueError(f"Ensemble {self.ensemble} not valid")
        ensemble_type = self.ensemble_specs.get("type", default_ensemble_types[self.ensemble])
        if ensemble_type not in _valid_dynamics[self.ensemble]:
            raise ValueError(
                f"Type {ensemble_type} is not valid for {self.ensemble} ensemble. "
                f"Valid types are: {' ,'.join(_valid_dynamics[self.ensemble])}"
            )
        ensemble_name = f"{self.ensemble.upper()}_{ensemble_type}" if ensemble_type else self.ensemble.upper()
        updates["MD_run"].append(ensemble_name)

        # add temperature
        if self.ensemble == "nve":
            if not self.init_velocities and "velocity" not in structure.site_properties:
                raise ValueError("Velocities must be initialized for NVE ensemble")
        else:
            if self.temp is None:
                raise ValueError(f"Temperature must be set for {ensemble_name} ensemble")
            updates["MD_run"].append(self.temp)

        # check for ensemble control parameter
        ensemble_parameter = self.ensemble_specs.get("parameter", None)
        if ensemble_name not in ("NVE", "NVE_4th_order"):
            if ensemble_parameter is None:
                raise ValueError(f"Ensemble {ensemble_name} parameter is not defined")
            updates["MD_run"].append(ensemble_parameter)

        # ...and put everything in the string
        updates["MD_run"] = " ".join(map(str, updates["MD_run"]))

        # initialize velocities
        if self.init_velocities:
            if self.temp is None:
                raise ValueError("Temperature must be set for velocity initialisation")
            updates["MD_MB_init"] = self.temp

        return updates



"""Parsers for FHI-aims standard output file."""

import numpy as np

from . import converters
from .abc import ChunkParser, FileParser


class RootParser(FileParser):
    """Base parser for output parsing."""

    name = "root"
    initial_chunk = "preamble"


class PreambleParser(ChunkParser):
    """Parse the preamble portion of aims.out file."""

    name = "preamble"
    values = {
        "aims_version": r"FHI-aims version *: (\w*)",
        "start_date": converters.to_date(r"Date *:  (\d*)"),
        "start_time": converters.to_time(r"Time *:  ([\d.]*)"),
    }
    metadata = {
        "aims_version": r"FHI-aims version *: (\w*)",
        "commit_hash": r"Commit number *: (\w*)",
        "aims_uuid": r"aims_uuid : ([^\n]+)$",
        "fortran_compiler": r"Fortran compiler *: ([^\n]+)$",
        "fortran_compiler_flags": r"Fortran compiler flags: ([^\n]+)$",
        "c_compiler": r"C compiler *: ([^\n]+)$",
        "c_compiler_flags": r"C compiler flags *: ([^\n]+)$",
        "cxx_compiler": r"C\+\+ compiler *: ([^\n]+)$",
        "cxx_compiler_flags": r"C\+\+ compiler flags *: ([^\n]+)$",
        "elpa_kernel": r"  ELPA2 kernel *: (\w*)",
        "build_type": converters.to_vector(r"Using (\S+)", dtype=str, multistring=True),
        "linked_against": converters.to_vector(
            r"(?:Linking against:|^ *) (\/\S+)", dtype=str, multistring=True
        ),
        "num_tasks": converters.to_int(r"Using *(\d+) parallel tasks."),
    }
    next_chunks = ["control_in"]
    parsed_key = ""


class ControlInParser(ChunkParser):
    """Parse the control.in portion of aims.out file."""

    name = "control_in"
    title_line = r"Parsing control.in"
    values = {"control_in": r"-{71}\n(.+)  -{71}\n  Completed first pass"}
    next_chunks = ["geometry_in"]
    parsed_key = "input"

    def parse(self):
        """Make additional adjustments to parse results."""
        parse_results = super().parse()
        # get the runtime choices
        choices = {}
        if "control_in" not in parse_results:
            return parse_results
        for line in parse_results["control_in"].split("\n"):
            if "species" in line:
                break
            hash_idx = line.find("#")
            keywords = (line[:hash_idx] if hash_idx >= 0 else line).strip().split()
            if keywords:
                choices[keywords[0]] = " ".join(keywords[1:])
        self.parent.run_metadata.update(choices)
        return parse_results


class GeometryInParser(ChunkParser):
    """Parse the geometry.in portion of aims.out file."""

    name = "geometry_in"
    title_line = r"Parsing geometry.in"
    values = {"geometry_in": r"-{71}\n(.+)  -{71}\n  Completed first pass"}
    metadata = {
        "num_species": converters.to_int(r"Number of species *: *(\d+)"),
        "num_atoms": converters.to_int(r"Number of atoms *: *(\d+)"),
        "num_spins": converters.to_int(r"Number of spin channels *: *(\d+)"),
    }
    next_chunks = ["reading_control_in"]
    parsed_key = "input"


class ReadingControlParser(ChunkParser):
    """Parse the reading control portion of aims.out file."""

    name = "reading_control_in"
    title_line = r"Reading file control.in"
    values = {
        "electronic_temperature": converters.to_float(
            r"broadening, width = *([-+.E\d]+) eV."
        )
    }
    next_chunks = ["reading_geometry_in"]
    parsed_key = "input"
    metadata = {
        "relax": converters.to_bool(r"Geometry relaxation"),
        "md": converters.to_bool(r"Molecular dynamics"),
    }


class ReadingGeometryParser(ChunkParser):
    """Parse the reading geometry portion of aims.out file."""

    name = "reading_geometry_in"
    title_line = r"Reading geometry description geometry.in"
    values = {
        "geometry": converters.to_atoms(r"Input geometry:\n(.*?)$^\n", from_input=True)
    }
    next_chunks = ["consistency_checks"]
    parsed_key = "input"


class ConsistencyChecksParser(ChunkParser):
    """Parse the consistency check portion of aims.out file."""

    name = "consistency_checks"
    title_line = r"Consistency checks .* are next"
    values = {
        "electronic_temperature": converters.to_float(
            r"broadening, width = *([-+.E\d]+) eV."
        )
    }
    metadata = {
        "num_bands": converters.to_int(
            r"Number of Kohn-Sham states \(occupied \+ empty\): *(\d+)"
        ),
    }
    next_chunks = ["fixed_parts"]
    parsed_key = "input"


class FixedPartsParser(ChunkParser):
    """Parse the fixed values portion of aims.out file."""

    name = "fixed_parts"
    title_line = r"Preparing all fixed parts"
    values = {
        "k_points": converters.to_matrix(
            r"(?:  \| k-point: *\d+ at *([-.\d]+) * *([-.\d]+) *([-.\d]+) *, "
            r"weight: *([-.\d]+))+",
            dtype=[float, float, float, float],
        )
    }
    metadata = {
        "num_bands": converters.to_int(
            r"Reducing total number of +Kohn-Sham states to *(\d+)"
        ),
        "num_basis_fns": converters.to_int(r"Total number of basis functions : *(\d+)"),
        "num_k_points": converters.to_int(r"Number of k-points *: *(\d+)"),
    }
    next_chunks = ["scf_init"]
    parsed_key = "input"

    def parse(self) -> dict:
        """Make additional adjustments to fixed-parts parse results."""
        parse_results = super().parse()
        if "k_points" in parse_results:
            parse_results.update(
                {
                    "k_points": [p[:3] for p in parse_results["k_points"]],
                    "k_point_weights": [p[3] for p in parse_results["k_points"]],
                }
            )
            assert len(parse_results["k_points"]) == self.run_metadata["num_k_points"]
        return parse_results


class SCFInitParser(ChunkParser):
    """Parse the SCF Initiation portion of aims.out file."""

    name = "scf_init"
    title_line = r"Begin self-consistency loop"
    next_chunks = ["init_values"]
    parsed_key = "ionic_steps[]"


class ScfInitValuesParser(ChunkParser):
    """Parse the SCF initialization portion of aims.out file."""

    name = "init_values"
    title_line = r"-{60}"
    values = {
        "chemical_potential": converters.to_float(
            r"  \| Chemical potential \(Fermi level\): *([-+.E\d]*) eV"
        ),
        "vbm": converters.to_float(r"Highest occupied state \(VBM\) at *([-.\d]*) eV"),
        "cbm": converters.to_float(r"Lowest unoccupied state \(CBM\) at *([-.\d]*) eV"),
        "gap": converters.to_float(r"HOMO-LUMO gap: *([-+.E\d]*) eV"),
        "total_energy": converters.to_float(
            r"  \| Total energy *: *[-.\d]* Ha *([-.\d]*) eV"
        ),
        "free_energy": converters.to_float(
            r"  \| Electronic free energy *: *[-.\d]* Ha *([-.\d]*) eV"
        ),
    }
    metadata = {
        "num_electrons": converters.to_float(
            r"Formal number of electrons \(from input files\) : *([\d.]+)"
        ),
        "num_k_points": converters.to_int(r"Number of k-points *: *(\d+)"),
    }
    next_chunks = [
        {"runtime_choices": {"output_level": "MD_light"}, "chunk": "mdlight_scf"},
        "scf_init",
        "scf_step",
        "final_values",
    ]
    parsed_key = "ionic_steps.scf_init"


class SCFStepParser(ChunkParser):
    """Parse the SCF step portion of aims.out file."""

    name = "scf_step"
    title_line = r"Begin self-consistency iteration"
    values = {
        "chemical_potential": converters.to_float(
            r"  \| Chemical potential \(Fermi level\): *([-+.E\d]*) eV"
        ),
        "vbm": converters.to_float(r"Highest occupied state \(VBM\) at *([-.\d]*) eV"),
        "cbm": converters.to_float(r"Lowest unoccupied state \(CBM\) at *([-.\d]*) eV"),
        "gap": converters.to_float(r"HOMO-LUMO gap: *([-+.E\d]*) eV"),
        "direct_gap": converters.to_float(r"Smallest direct gap : *([-+.E\d]*) eV"),
        "is_metallic": converters.to_bool(r"this material is metallic"),
        "spin": converters.to_float(
            r"  \| N = N_up - N_down (?:\(sum over all k points\)|): *([-.\d]*)"
        ),
        "total_energy": converters.to_float(
            r"  \| Total energy *: *[-.\d]* Ha *([-.\d]*) eV"
        ),
        "free_energy": converters.to_float(
            r"  \| Electronic free energy *: *[-.\d]* Ha *([-.\d]*) eV"
        ),
        "charge_density_change": converters.to_float(
            r"  \| Change of charge density *: *([-+.E\d]*)"
        ),
        "charge_spin_density_change": converters.to_vector(
            r"  \| Change of charge/spin density *: *([-+.E\d]*)  *([-+.E\d]*)",
            dtype=float,
        ),
        "eigenvalues_sum_change": converters.to_float(
            r"  \| Change of sum of eigenvalues *: *([-+.E\d]*)"
        ),
        "total_energy_change": converters.to_float(
            r"  \| Change of total energy *: *([-+.E\d]*)"
        ),
        "change_of_forces": converters.to_float(
            r"  \| Change of forces *: *([-+.E\d]*) eV/A"
        ),
        "analytical_stress": converters.to_table(
            r"  \| *Analytical stress tensor - Symmetrized *\|",
            header=5,
            num_rows=3,
            dtype=[None, None, float, float, float, None],
        ),
        "pressure": converters.to_float(r"Pressure: *([-+.E\d]*)"),
        "charge_density_status": r"Charge density*: ([^\n]+)$",
        "spin_density_status": r"Spin density*: ([^\n]+)$",
        "estimated_steps_to_convergence": converters.to_int(
            r"Estimated *(\d+)* more steps to reach target accuracy"
        ),
    }
    next_chunks = ["scf_step", "converged_scf", "not_converged_scf", "final_values"]
    parsed_key = "ionic_steps.scf_steps[]"

    def parse(self) -> dict:
        """Make additional adjustments to parse results."""
        parse_results = super().parse()
        if "charge_spin_density_change" in parse_results:
            charge, spin = parse_results.pop("charge_spin_density_change")
            parse_results["charge_density_change"] = charge
            parse_results["spin_density_change"] = spin

        if "charge_density_status" in parse_results:
            if "estimated_steps_to_convergence" in parse_results:
                parse_results["convergence_expected"] = True
            else:
                parse_results["convergence_expected"] = False

        return parse_results


class MDLightSCFParser(ChunkParser):
    """Parse the SCF steps table of aims.out file for `MD_light` output level."""

    name = "mdlight_scf"
    title_line = r"Convergence:"
    next_chunks = ["mdlight_energy"]
    parsed_key = "ionic_steps"

    def parse(self):
        """Make additional adjustments to parse results."""
        content = self.collect()
        step_lines = []
        for line in content.splitlines()[1:-1]:
            if line.startswith("  SCF"):
                step_lines.append(line)
            else:
                step_lines[-1] += line
        # Each step:  q app. | density  | eigen (eV) | Etot (eV) | forces (eV/A) | CPU time | Clock time # noqa: E501
        steps = []
        for step in step_lines:
            results = [val for val in step.split("|") if not val.strip()[0].isalpha()]

            steps.append(
                {
                    "eigenvalues_sum_change": float(results[1]),
                    "total_energy_change": float(results[2]),
                    "cpu_time": float(results[4].split()[0]),
                    "wall_time": float(results[5].split()[0]),
                }
            )
            try:
                charge_spin_change = results[0].split()
                steps[-1]["charge_density_change"] = float(charge_spin_change[0])
                if len(charge_spin_change) > 1:
                    steps[-1]["spin_density_change"] = float(charge_spin_change[1])
                steps[-1]["change_of_forces"] = float(results[3])
            except ValueError:
                pass
        return {"scf_steps": steps}


class MDLightEnergyParser(SCFStepParser):
    """Parse the SCF energy for `MD_light` output level of aims.out file."""

    name = "mdlight_energy"
    title_line = r"Total energy components:"
    values = {
        "chemical_potential": converters.to_float(
            r"  \| Chemical potential \(Fermi level\): *([-+.E\d]*) eV"
        ),
        "vbm": converters.to_float(r"Highest occupied state \(VBM\) at *([-.\d]*) eV"),
        "cbm": converters.to_float(r"Lowest unoccupied state \(CBM\) at *([-.\d]*) eV"),
        "gap": converters.to_float(r"HOMO-LUMO gap: *([-+.E\d]*) eV"),
        "direct_gap": converters.to_float(r"Smallest direct gap : *([-+.E\d]*) eV"),
        "is_metallic": converters.to_bool(r"this material is metallic"),
        "spin": converters.to_float(
            r"  \| N = N_up - N_down (?:\(sum over all k points\)|): *([-.\d]*)"
        ),
        "total_energy": converters.to_float(
            r"  \| Total energy *: *[-.\d]* Ha *([-.\d]*) eV"
        ),
        "free_energy": converters.to_float(
            r"  \| Electronic free energy *: *[-.\d]* Ha *([-.\d]*) eV"
        ),
    }
    next_chunks = [
        "scf_init",
        "converged_scf",
        "mulliken",
        "final_values",
    ]
    parsed_key = "ionic_steps"


class NotConvergedSCFParser(ChunkParser):
    """Parse the non-converged SCF part of aims.out file."""

    name = "not_converged_scf"
    title_line = r"WARNING! SELF-CONSISTENCY CYCLE DID NOT CONVERGE"
    values = {
        "scf_converged": converters.to_bool(r"Self-consistency cycle converged"),
    }
    next_chunks = [
        "hirshfeld",
        "mulliken",
        "final_values",
    ]
    parsed_key = "ionic_steps"


class ConvergedSCFParser(ChunkParser):
    """Parse the converged SCF values of aims.out file."""

    name = "converged_scf"
    title_line = r"Self-consistency cycle converged"
    values = {
        "scf_converged": converters.to_bool(r"Self-consistency cycle converged"),
        "total_energy": converters.to_float(
            r"  \| Total energy uncorrected *: *([-+.E\d]*) eV"
        ),
        "corrected_total_energy": converters.to_float(
            r"  \| Total energy corrected *: *([-+.E\d]*) eV"
        ),
        "free_energy": converters.to_float(
            r"  \| Electronic free energy *: *([-+.E\d]*) eV"
        ),
        "vdw_correction": converters.to_float(
            r"  \| vdW energy correction *: *[-.\d]* Ha *([-.\d]*) eV"
        ),
        # charge / dipole moments
        "total_charge": converters.to_float(r"  \| Total charge \[e\] *: *([-+.E\d]*)"),
        "total_dipole_moment": converters.to_vector(
            r"  \| Total dipole moment \[eAng\] *: "
            r"*([-+.E\d]*) *([-+.E\d]*) *([-+.E\d]*)",
            dtype=float,
        ),
        "absolute_dipole_moment": converters.to_float(
            r"  \| Absolute dipole moment *: *([-+.E\d]*)"
        ),
        "hirshfeld_charges": converters.to_vector(
            r"  \|   Hirshfeld charge *: *([-+.\d]*)", multistring=True
        ),
        # forces and stress
        "atomic_forces": converters.to_table(
            r"Total atomic forces [^\n]* \[eV\/Ang\]:",
            num_rows="num_atoms",
            dtype=[None, None, float, float, float],
        ),
        "analytical_stress": converters.to_table(
            r"  \| *Analytical stress tensor - Symmetrized *\|",
            header=5,
            num_rows=3,
            dtype=[None, None, float, float, float, None],
        ),
        "numerical_stress": converters.to_table(
            r"  \| *Numerical stress tensor *\|",
            header=5,
            num_rows=3,
            dtype=[None, None, float, float, float, None],
        ),
        "pressure": converters.to_float(r"Pressure: *([-+.E\d]*)"),
        "atomic_stresses": converters.to_table(
            r"Per atom stress \(eV\) used for heat flux calculation:",
            header=3,
            num_rows="num_atoms",
            dtype=[None, None, float, float, float, float, float, float],
        ),
    }
    next_chunks = [
        "scf_init",
        "ionic_step_geometry",
        {
            "runtime_choices": {"calculate_atom_bsse": True},
            "chunk": "bsse_ionic_step_geometry",
        },
        {"runtime_choices": {"md": True}, "chunk": "md_values"},
        {"runtime_choices": {"qpe_calc": "gw_expt"}, "chunk": "periodic_gw"},
        "final_geometry",
        "mulliken",
        "hirshfeld",
        "output_polarization",
        "dfpt_dielectric",
        "final_values",
    ]
    parsed_key = "ionic_steps"


class MullikenAnalysisParser(ChunkParser):
    """Parse the Mulliken analysis part of aims.out file."""

    name = "mulliken"
    title_line = r" *Starting Mulliken Analysis"
    values = {
        "mulliken_charges": converters.to_table(
            r"Performing (?:scalar-relativistic )?Mulliken charge[\s\S]*?"
            r"Summary of the per-atom charge analysis:",
            header=3,
            num_rows="num_atoms",
            dtype=[None, None, None, float, None, None, None, None],
        ),
        "mulliken_spins": converters.to_table(
            r"Performing (?:scalar-relativistic )?Mulliken charge[\s\S]*?"
            r"Summary of the per-atom spin analysis:",
            header=3,
            num_rows="num_atoms",
            dtype=[None, None, float, None, None, None],
        ),
        "mulliken_charges_soc": converters.to_table(
            r"Performing spin-orbit-coupled Mulliken charge[\s\S]*?"
            r"Summary of the per-atom charge analysis:",
            header=3,
            num_rows="num_atoms",
            dtype=[None, None, None, float, None, None, None, None],
        ),
        "mulliken_spins_soc": converters.to_table(
            r"Performing spin-orbit-coupled Mulliken charge[\s\S]*?"
            r"Summary of the per-atom spin analysis:",
            header=3,
            num_rows="num_atoms",
            dtype=[None, None, float, None, None, None],
        ),
    }
    next_chunks = [
        "hirshfeld",
        "output_polarization",
        "dfpt_dielectric",
        "final_values",
    ]
    parsed_key = "final"


class DFPTDielectricParser(ChunkParser):
    """Parse the DFPT analysis part of aims.out file."""

    name = "dfpt_dielectric"
    title_line = r"ENTERING DFPT_DIELECTRIC"
    values = {
        "polarizability": converters.to_vector(
            r"  \| Polarizability:---> *([-+.\d]*) *([-+.\d]*) *([-+.\d]*) *([-+.\d]*) *([-+.\d]*) *([-+.\d]*)",  # noqa: E501
            dtype=float,
        ),
        "dielectric_tensor": converters.to_table(
            r"DFPT for dielectric_constant:--->  # PARSE DFPT_dielectric_tensor",
            header=1,
            num_rows=3,
            dtype=[float, float, float],
        ),
    }
    next_chunks = ["output_polarization", "final_values"]
    parsed_key = "dfpt"


class PolarizationParser(ChunkParser):
    """Parse the Polarization calculation part of aims.out file."""

    name = "output_polarization"
    title_line = r"Starting with the calculation of the polarization"
    values = {
        "polarization": converters.to_vector(
            r"  \| Cartesian Polarization *([-+.\d]*) *([-+.\d]*) *([-+.\d]*)",
            dtype=float,
        ),
    }
    next_chunks = ["dfpt_dielectric", "final_values"]
    parsed_key = "polarization"


class HirshfeldAnalysisParser(ChunkParser):
    """Parse the Hirshfeld analysis part of aims.out file."""

    name = "hirshfeld"
    title_line = r"Performing Hirshfeld analysis of fragment charges and moments."
    values = {
        "hirshfeld_charges": converters.to_vector(
            r"  \|   Hirshfeld charge *: *([-+.\d]*)", multistring=True
        ),
        "hirshfeld_spins": converters.to_vector(
            r"  \|   Hirshfeld spin moment *: *([-+.\d]*)", multistring=True
        ),
        "hirshfeld_dipoles": converters.to_vector(
            r"  \|   Hirshfeld dipole moment *: *([-+.\d]*)", multistring=True
        ),
    }
    next_chunks = ["final_values"]
    parsed_key = "final"


class PeriodicGWParser(ChunkParser):
    """Parse the Periodic GW portion of aims.out file."""

    name = "periodic_gw"
    title_line = r"Initializing LVL tricoefficents in reciprocal space for GW ..."
    values = {
        "vbm": converters.to_float(
            r"\"GW Band gap\" of total set of bands:[\s\S]*?"
            r"  \| Highest occupied state : *([-.\d]*) eV"
        ),
        "cbm": converters.to_float(
            r"\"GW Band gap\" of total set of bands:[\s\S]*?"
            r"  \| Lowest unoccupied state: *([-.\d]*) eV"
        ),
        "gap": converters.to_float(
            r"\"GW Band gap\" of total set of bands:[\s\S]*?"
            r"  \| Energy difference      : *([-.\d]*) eV"
        ),
        "se_on_k_grid": r" *GW quasi-particle energy levels\n"
        r"(?:[^\n]*\n){5}([\s\S]*)\n\n"
        r"  Valence band maximum",
        "se_states": converters.to_vector(
            r"states to compute self-energy: *([\d]*) *([\d]*)", dtype=int
        ),
    }
    next_chunks = ["final_values"]
    parsed_key = "gw"

    def parse(self) -> dict:
        """Additionally, parse the table of the self-energy
        on the regular k-point grid.
        """
        parse_results = super().parse()
        # parse the table the usual way
        try:
            se_str = parse_results.pop("se_on_k_grid")
            states = parse_results.pop("se_states")
        except KeyError:
            return parse_results
        se_lines = se_str.split("\n")
        headers = se_lines[0].split()[1:]
        n_states = states[1] - states[0] + 1
        n_k_points = (len(se_lines) - 1) / (n_states + 4)
        se = {
            "states": list(range(states[0], states[1] + 1)),
        }
        result = []
        for i_k in range(int(n_k_points)):
            result.append([])
            for i_line in range(i_k * (n_states + 4) + 4, (i_k + 1) * (n_states + 4)):
                result[i_k].append(list(map(float, se_lines[i_line].split()[1:])))
        result = np.array(result)
        for i_head, head in enumerate(headers):
            se[head] = result[:, :, i_head].T
        parse_results["self_energy"] = se
        return parse_results


class MDValuesParser(ChunkParser):
    """Parse the MD porition of aims.out file."""

    name = "md_values"
    title_line = r"  Complete information for previous time-step:"
    values = {
        "time_step": converters.to_int(r"  \| Time step number *: *(\d+)"),
        "electronic_free_energy": converters.to_float(
            r"  \| Electronic free energy *: *([-+.E\d]*) eV"
        ),
        "temperature": converters.to_float(
            r"  \| Temperature \(nuclei\) *: *([-+.E\d]*) K"
        ),
        "kinetic_energy": converters.to_float(
            r"  \| Nuclear kinetic energy *: *([-+.E\d]*) eV"
        ),
        "total_energy": converters.to_float(
            r"  \| Total energy \(el.+nuc.\) *: *([-+.E\d]*) eV"
        ),
        "gle_H": converters.to_float(r"  \| GLE pseudo hamiltonian *: *([-+.E\d]*) eV"),
    }
    next_chunks = ["ionic_step_geometry"]
    parsed_key = "ionic_steps.md"


class IonicStepGeometry(ChunkParser):
    """Parse the geometry of ionic step in an aims.out file."""

    name = "ionic_step_geometry"
    title_line = (
        r"Atomic structure \(and velocities\) as used in the preceding time step:|"
        r"Final atomic structure \(and velocities\) as used in the preceding time step:|"  # noqa: E501
        r"Atomic structure that was used in the preceding time step of the wrapper:|"
        r"Updated atomic structure:"
    )
    values = {
        "geometry": converters.to_atoms(
            r"Atomic structure \(and velocities\) as used in the preceding time step:|"
            r"Final atomic structure \(and velocities\) as used in the preceding time step:|"  # noqa: E501
            r"Atomic structure that was used in the preceding time step of the wrapper|"
            r"Updated atomic structure:"
        )
    }
    next_chunks = ["scf_init", "mulliken", "final_values"]
    parsed_key = "ionic_steps"


class BSSEIonicStepGeometry(IonicStepGeometry):
    """Parse the geometry of BSSE step in an aims.out file."""

    name = "bsse_ionic_step_geometry"
    title_line = r"Geometry for the next BSSE step"
    values = {"geometry": converters.bsse_to_atoms()}


class FinalGeometry(ChunkParser):
    """Parse the final geometry in a aims.out file."""

    name = "final_geometry"
    title_line = r"Present geometry is converged|Aborting optimization"
    values = {
        "geometry_converged": converters.to_bool(r"Present geometry is converged"),
        "geometry": converters.to_atoms(r"  Final atomic structure:"),
    }
    next_chunks = ["mulliken", "output_polarization", "dfpt_dielectric", "final_values"]
    parsed_key = "final"


class FinalValuesParser(ChunkParser):
    """Parse the final physical values in an aims.out file."""

    name = "final_values"
    title_line = r"Final output of selected total energy values"
    values = {
        "energy": converters.to_float(
            r"  \| Total energy of the DFT / Hartree-Fock s.c.f. "
            r"calculation *: *([-.\d]*)"
        ),
    }
    parsed_key = "final"
    next_chunks = ["final"]


class FinalParser(ChunkParser):
    """Parse the final metadata values in an aims.out file."""

    name = "final"
    title_line = r"Leaving FHI-aims"
    values = {
        "end_date": converters.to_date(r"Date *:  (\d*)"),
        "end_time": converters.to_time(r"Time *:  ([\d.]*)"),
        "num_scf_steps": converters.to_int(
            r"Number of self-consistency cycles *: *(\d+)"
        ),
        "num_ionic_steps": converters.to_int(
            r"Number of SCF \(re\)initializations *: *(\d+)"
        ),
        "num_relax_steps": converters.to_int(r"Number of relaxation steps *: *(\d+)"),
        "num_md_steps": converters.to_int(
            r"Number of molecular dynamics steps *: *(\d+)"
        ),
        "num_force_evals": converters.to_int(r"Number of force evaluations *: *(\d+)"),
    }
    next_chunks = ["final_times"]
    parsed_key = ""


class FinalTimesParser(ChunkParser):
    """Parse the final timing values in an aims.out file."""

    name = "final_times"
    title_line = r"Detailed time accounting"
    values = {
        "total": converters.to_vector(r"Total time *: *([\d.]+) s *([\d.]+) s"),
        "preparation": converters.to_vector(
            r"Preparation time *: *([\d.]+) s *([\d.]+) s"
        ),
        "bc_init": converters.to_vector(
            r"Boundary condition initalization *: *([\d.]+) s *([\d.]+) s"
        ),
        "grid_part": converters.to_vector(
            r"Grid partitioning *: *([\d.]+) s *([\d.]+) s"
        ),
        "preloading_free_atom": converters.to_vector(
            r"Preloading free-atom quantities on grid *: *([\d.]+) s *([\d.]+) s"
        ),
        "free_atom_sp_e": converters.to_vector(
            r"Free-atom superposition energy *: *([\d.]+) s *([\d.]+) s"
        ),
        "integration": converters.to_vector(
            r"Total time for integrations *: *([\d.]+) s *([\d.]+) s"
        ),
        "ks_equations": converters.to_vector(
            r"Total time for solution of K.-S. equations *: *([\d.]+) s *([\d.]+) s"
        ),
        "ev_reortho": converters.to_vector(
            r"Total time for EV reorthonormalization *: *([\d.]+) s *([\d.]+) s"
        ),
        "density_force": converters.to_vector(
            r"Total time for density & force components *: *([\d.]+) s *([\d.]+) s"
        ),
        "mixing": converters.to_vector(
            r"Total time for mixing *: *([\d.]+) s *([\d.]+) s"
        ),
        "hartree_update": converters.to_vector(
            r"Total time for Hartree multipole update *: *([\d.]+) s *([\d.]+) s"
        ),
        "hartree_sum": converters.to_vector(
            r"Total time for Hartree multipole sum *: *([\d.]+) s *([\d.]+) s"
        ),
        "total_energy_eval": converters.to_vector(
            r"Total time for total energy evaluation *: *([\d.]+) s *([\d.]+) s"
        ),
        "nsc_force": converters.to_vector(
            r"Total time NSC force correction *: *([\d.]+) s *([\d.]+) s"
        ),
        "scaled_zora": converters.to_vector(
            r"Total time for scaled ZORA corrections *: *([\d.]+) s *([\d.]+) s"
        ),
        "pert_soc": converters.to_vector(
            r"Total time for perturbative SOC *: *([\d.]+) s *([\d.]+) s"
        ),
        "wannier_evol": converters.to_vector(
            r"Total time for Wannier Center Evolution *: *([\d.]+) s *([\d.]+) s"
        ),
    }
    next_chunks = ["have_a_nice_day"]
    parsed_key = "time"


class HaveANiceDayParser(ChunkParser):
    """Parse if the aims.out file exited properly."""

    name = "have_a_nice_day"
    title_line = r"Have a nice day"
    values = {"is_finished_ok": converters.to_bool(r"(Have a nice day.)")}
    parsed_key = ""



"""A main output file parser for FHI-aims.

Realizes the Deterministic State Machine pattern.
"""

from __future__ import annotations

import importlib
import inspect
import random
import re
import string
from abc import abstractmethod
from collections.abc import Callable
from pathlib import Path
from typing import IO, Any, ClassVar

from monty.io import zopen

from .converters import to_str
from .utils import update

INT = r"[-\d]+"
FLOAT = r"[-+.\d]+"
EXP = r"[-+.E\d]+"


def parser_classes(module) -> dict[str, type[ChunkParser]]:
    """Return the dict of all the file and chunk parser classes declared in
    the current module.
    """
    result = {}
    for _, cls in inspect.getmembers(module, inspect.isclass):
        if issubclass(cls, ChunkParser):
            result[cls.name] = cls
    return result


def _get_parser(parser_class, header_line):
    """Inject the header line into the ChunkParser instance."""

    def _factory(file_descriptor: str | Path | IO, parent: Parser = None) -> Parser:
        return parser_class(file_descriptor, parent, header_line)

    return _factory


class Parser:
    """Abstract parser class for FHI-aims output files."""

    name: str = ""

    def __init__(
        self, file_descriptor: str | Path | IO, parent: FileParser = None
    ) -> None:
        """Construct a base class for the parser.

        Parameters
        ----------
        file_descriptor: str | Path | IO
            A name or handle of the main output file
        parent: Parser
            A parent parser for the current instance

        """
        # get the module in which the actual implementation is located
        # (parsers.py) and available parsers within that module
        module = importlib.import_module(self.__module__)
        self._available_parsers = parser_classes(module)
        self.result = {}
        self.warnings = []
        self.errors = []
        # propagate metadata along inheritance line
        self.parent = parent
        if self.parent is None:
            self.run_metadata = {}
        else:
            self.run_metadata = self.parent.run_metadata
        self.next = None
        # open the file to parse
        if isinstance(file_descriptor, str | Path):
            self.fd = zopen(file_descriptor, "rt")
            self.is_root = True
        else:
            self.fd = file_descriptor
            self.is_root = False

    def add_parsed_values(self, parser_name: str, **kwargs):
        """Update the specified parser's values with the provided key-value pairs.
        Only parsers available in the system can be updated. Each value should either
        be of type string or returned from a function from a `converters` module.
        Raises an error if the parser is not available or if a value is of an
        unsupported type.

        Args:
            parser_name: The name of the parser to be updated.
            kwargs: Key-value pairs to be added to the parser's values.

        Raises:
            ValueError: If the parser is not available
            TypeError: If any of the provided values are not of a valid type.

        """
        parser = self._available_parsers.get(parser_name, None)
        if parser is None:
            raise ValueError(f"Parser {parser_name} is not available.")
        for v in kwargs.values():
            if not isinstance(v, str | Callable):
                raise TypeError(
                    f"The parsed value {v} should either be a string or returned from "
                    f"`converters.to_...` function. Got {type(v)} instead.`."
                )
        parser.values.update(kwargs)

    def __del__(self):
        """Close the file if it is the root parser."""
        if self.is_root:
            self.fd.close()

    @abstractmethod
    def parse(self):
        """Parse the output file by dividing it into chunks.

        It then parses each chunk separately.

        Returns
        -------
        dict: a parsed dictionary

        """
        raise NotImplementedError

    @abstractmethod
    def annotate(self, annotated_file: str | Path | IO):
        """Annotates the output file by adding the chunk name and ID to each line.

        Parameters
        ----------
        annotated_file: str | Path | IO
            The name of the file to write annotations to.
            If None, write the annotated file to stdout.

        """
        raise NotImplementedError


class FileParser(Parser):
    """The main output file parser."""

    initial_chunk: str

    def __init__(
        self,
        file_descriptor: str | Path | IO,
        parent: Parser = None,
        header_line: str | None = None,
    ) -> None:
        """Construct the main Parser object.

        Parameters
        ----------
        file_descriptor : str | Path | IO
            A name or handle of the main output file
        parent: Parser
            The parent parser
        header_line: str | None
            The first line to serach for

        """
        super().__init__(file_descriptor, parent)
        # inject header line into the ChunkParser class
        self._initial_chunk = _get_parser(
            self._available_parsers[self.initial_chunk], header_line
        )

    def parse(self) -> dict:
        """Parse the output file by dividing it into chunks.

        It then parses each chunk separately.

        Returns
        -------
        dict: a parsed dictionary

        """
        self.fd.seek(0)
        # instantiate ChunkParser with header line already injected
        current_chunk = self._initial_chunk(self.fd, parent=self)
        # can we create the `parse_results` structure preliminary?
        parse_results = {}
        while current_chunk:
            chunk_results = current_chunk.parse()
            if chunk_results or current_chunk.parsed_key:
                update(parse_results, chunk_results, current_chunk.parsed_key)
            if current_chunk.next is None:
                break
            current_chunk = current_chunk.next(self.fd, parent=self)
        self.result = parse_results
        return self.result

    def annotate(self, annotated_file: str | Path | IO | None = None) -> None:
        """Annotates the output file by adding the chunk name and ID to each line.

        Parameters
        ----------
        annotated_file: str | Path | IO
            The name of the file to write annotations to.
            If None, write the annotated file to stdout.

        """
        self.fd.seek(0)
        if isinstance(annotated_file, str | Path):
            annotated_fd = open(annotated_file, "w")  # noqa: SIM115
        else:
            annotated_fd = annotated_file
        current_chunk = self._initial_chunk(self.fd, parent=self)
        while current_chunk:
            current_chunk.annotate(annotated_fd)
            if current_chunk.next is None:
                break
            current_chunk = current_chunk.next(self.fd, parent=self)
        if isinstance(annotated_fd, IO):
            annotated_fd.close()


class ChunkParser(Parser):
    """An abstract class for the output file chunks.

    Attributes
    ----------
    name : str
        the name of the chunk parser
    title_line : str
        The start line of the given chunk in the output file
    values : dict[str, Any]
        the dict of the names and regular expressions for the values that
        may be found in the chunk
    next_chunks : list[str | dict]
        the collection of lines and runtime choices that define the next chunk
    parsed_key: str
        the key under which the values are stored in the parse results

    Parameters
    ----------
    file_descriptor : str | Path | IO
        a path to the file or an open output file descriptor

    """

    title_line: str = ""
    parsed_key: str = ""
    values: ClassVar[dict[str, Any]] = {}
    metadata: ClassVar[dict[str, Any]] = {}
    next_chunks: ClassVar[list[str | dict[str, Any]]] = []

    def __init__(
        self,
        file_descriptor: str | Path | IO,
        parent: Parser = None,
        header_line: str | None = None,
    ) -> None:
        """Concstruct the parser.

        Parameters
        ----------
        file_descriptor: str | Path | IO
            The output file
        parent: Parser
            The parent parser
        header_line: str | None
            The first line to serach for

        """
        super().__init__(file_descriptor, parent)
        self.header_line = header_line
        self.uuid = "".join(random.choices(string.ascii_letters + string.digits, k=4))
        self._next_parsers = self._get_next_parsers()

    def _get_next_parsers(self) -> list[dict[str, Any]]:
        """Return the list of dictionaries containing the available next parsers
        and corresponding title lines.
        """
        next_parsers = []
        for chunk_meta in self.next_chunks:
            if isinstance(chunk_meta, str):
                next_parsers.append(
                    {
                        "parser": self._available_parsers[chunk_meta],
                        "line": self._available_parsers[chunk_meta].title_line,
                    }
                )
            elif isinstance(chunk_meta, dict) and "runtime_choices" in chunk_meta:
                parser_choices = chunk_meta.get("runtime_choices", {})
                on_parser = True
                for choice, value in parser_choices.items():
                    # boolean: check only for the existence of the key
                    if (
                        isinstance(value, bool)
                        and ((choice in self.run_metadata) != value)
                    ) or (
                        not isinstance(value, bool)
                        and self.run_metadata.get(choice, None) != value
                    ):
                        on_parser = False
                if on_parser:
                    next_parsers.append(
                        {
                            "parser": self._available_parsers[chunk_meta["chunk"]],
                            "line": self._available_parsers[
                                chunk_meta["chunk"]
                            ].title_line,
                        }
                    )
        return next_parsers

    def _check_for_next(self, line):
        """Decides on the chunk which a given line of the output should
        belong to.

        Parameters
        ----------
        line : str
            a line of the output file

        Returns
        -------
            None if the line belongs to this chunk; a FileChunk class if it belongs
            to another chunk

        """
        # find the next parser by comparing the line with the given regex
        for parser_meta in self._next_parsers:
            line_regex = parser_meta["line"]
            if re.search(line_regex, line):
                return _get_parser(parser_meta.get("parser", None), header_line=line)
        return None

    def collect(self):
        """Collect all lines belonging to this chunk."""
        lines = [self.header_line] if self.header_line is not None else []
        while line := self.fd.readline():
            next_chunk = self._check_for_next(line)
            if next_chunk is not None:
                self.next = next_chunk
                break
            lines.append(line)

        # TODO: Here check if the output file is actually complete!
        return "".join(lines)

    def parse(self) -> dict:
        """Parse the output file chunk genericlly."""
        line = self.collect()
        parse_result = {}
        # 1. values parsing
        for k, v in self.values.items():
            if not inspect.isfunction(v):
                v = to_str(v)
            # bring metadata to converter function; useful for parser variables
            match = v(line, self.run_metadata)
            if match is not None:
                parse_result[k] = match  # the first parenthesized subgroup
        # 2. metadata parsing
        for k, v in self.metadata.items():
            if not inspect.isfunction(v):
                v = to_str(v)
            match = v(line, self.run_metadata)
            if match is not None:
                self.parent.run_metadata[k] = match  # the first parenthesized subgroup
        # 3. warnings and errors parsing (ones that begin with optional
        # space and asterisk)
        self.parent.errors += re.findall(r"^(?: ?\*.*\n)+[\s\n]*\Z", line, re.MULTILINE)
        self.parent.warnings += re.findall(r"^(?: ?\*.*\n)+", line, re.MULTILINE)
        self.result = parse_result
        return self.result

    def annotate(self, annotated_file: IO) -> None:
        """Annotates the given chunk."""
        line = self.collect()
        # metadata parsing
        for k, v in self.metadata.items():
            if not inspect.isfunction(v):
                v = to_str(v)
            match = v(line, self.run_metadata)
            if match is not None:
                self.parent.run_metadata[k] = match  # the first parenthesized subgroup
        # get annotation and annotate
        truncated_name = self.name if len(self.name) < 15 else self.name[:12] + "..."
        annotation = f"{truncated_name:>15s}:{self.uuid}"
        annotated_lines = [f"{annotation} -> {el}" for el in line.split("\n")]
        if isinstance(annotated_file, IO):
            annotated_file.write("\n".join(annotated_lines))
        else:
            print("\n".join(annotated_lines), end="")



"""Tests for BasisSet class"""

from textwrap import dedent

import pytest

from pyfhiaims.species_defaults.basis_function import HydroBasisFunction
from pyfhiaims.species_defaults.basis_set import BasisSet, TierNotFoundError


def test_basis_set():
    """Test basis sets."""
    basis_string = dedent(
        """\
    #  "First tier" - improvements: -1014.90 meV to -62.69 meV
     hydro 2 s 2.1
     hydro 2 p 3.5
    #  "Second tier" - improvements: -12.89 meV to -1.83 meV
         hydro 1 s 0.85
    #     hydro 2 p 3.7
    #     hydro 2 s 1.2
      for_aux    hydro 3 d 7
    #  "Third tier" - improvements: -0.25 meV to -0.12 meV
     gaussian 0 2
        16.0396781            0.3942297
         6.5038187            0.2499824
    # gaussian 0 2
    #    18.0396781            0.3942297
    #     9.5038187            0.2499824
    #     hydro 4 f 11.2
    #     hydro 3 p 4.8
    #     hydro 4 d 9
    #     hydro 3 s 3.2

    """
    )
    basis_set = BasisSet.from_string(basis_string)
    assert basis_set.n_tiers == 3
    assert len(basis_set.tier(1)) == 2
    assert len(basis_set.tier(2)) == 3
    assert len(basis_set.tier(2, enabled=True)) == 1
    assert len(basis_set.tier(2, enabled=True, aux=True)) == 1
    assert len(basis_set.tier(3, enabled=True)) == 1

    basis_set.activate_tier(2, 1)
    assert len(basis_set.tier(2, enabled=True)) == 2

    basis_set.deactivate_functions(1)
    assert len(basis_set.tier(2, enabled=True)) == 2
    assert len(basis_set.tier(3, enabled=True)) == 0

    basis_set.activate_functions(1)
    assert len(basis_set.tier(2, enabled=True)) == 3

    basis_set.deactivate_tier(2, "all", aux=True)
    assert len(basis_set.tier(2, enabled=True, aux=True)) == 0

    basis_set.activate_tier(3, "all")
    assert len(basis_set.tier(3, enabled=True)) == 6

    basis_set.deactivate_tier(3, 99)
    assert len(basis_set.tier(3, enabled=True)) == 0

    basis_set.deactivate_functions(99)
    assert len(basis_set.tier(1, enabled=True)) == 0

    basis_set.add_function(
        HydroBasisFunction(n=1, l="s", z_eff=2.4), tier=3, active=True, aux=False
    )
    basis_set.add_function(
        HydroBasisFunction(n=1, l="s", z_eff=1.4), active=True, aux=False
    )
    assert len(basis_set.tier(3, enabled=True)) == 1
    assert len(basis_set.tier(6, enabled=True)) == 1

    with pytest.raises(TierNotFoundError):
        _ = basis_set.tier(100)

    with pytest.raises(TierNotFoundError):
        _ = basis_set.activate_tier(100)

    with pytest.raises(TierNotFoundError):
        _ = basis_set.add_function(
            HydroBasisFunction(n=1, l="s", z_eff=1.4), tier=100, active=True, aux=False
        )



"""A main pyaims class representing Calculation."""

from pyfhiaims.aims_input import AimsInput


class AimsCalc:
    """An Aims calculation."""

    aims_input: AimsInput

    @classmethod
    def from_folder(
        cls, folder, aims_stdout: str = "aims.out", aims_stderr: str | None = None
    ):
        """Get Aims Calculation from a folder with files."""



"""The geometry representation for geometry.in files."""

from __future__ import annotations

import time
import warnings
from collections import Counter
from collections.abc import Sequence
from dataclasses import dataclass, field, fields
from pathlib import Path
from typing import Any, TextIO

import numpy as np

from pyfhiaims.geometry.atom import ATOMIC_SYMBOLS_TO_NUMBERS, FHIAimsAtom
from pyfhiaims.species_defaults.species import SpeciesDefaults
from pyfhiaims.utils.typecast import Matrix3D, Vector3D, to_matrix3d


class InvalidGeometryError(Exception):
    """Exception raised if there is a problem with the input geometry."""

    def __init__(self, message: str) -> None:
        """Initialize the error with the message, message."""
        self.message = message
        super().__init__(self.message)


@dataclass
class AimsGeometry:
    """Geometry for the structure to run the calculation for.

    Parameters
    ----------
    atoms: Sequence[FHIAimsAtom]
        The Atoms for the geometry
    lattice_vectors: Optional[Matrix3D]
        The initial set of lattice vectors for the structure
    lattice_constraints: Sequence[tuple[bool, bool, bool]]
        fix lattice constraints for each nucleus and direction
    species_dict: dict[str, SpeciesDefaults]
        Dictionary to get the species for each symbol in the structure
    hessian_block: Optional[Sequence[tuple[int, int, Matrix3D]]]
        In geometry.in, allows to specify a Hessian matrix explicitly, with one
        line for each 3×3 block.
        The option block consists of nine numbers in column-first
        (Fortran) order. The 3×3 block corresponding to j_atom, i_atom is initialized
        by the transposed of block. The Hessian matrix is input in units of eV/Å2.
    hessian_block_lv: Optional[Sequence[tuple[float, float, Matrix3D]]]
        hessian_block for lattice vectors
    hessian_block_lv_atoms: Optional[Sequence[tuple[float, float, Matrix3D]]]
        hessian_block for degrees of freedom between lattice vectors and atoms
    hessian_file: Optional[bool]
        Indicates that there exists a hessian.aims file to be used to construct
        the Hessian.
    trust_radius: Optional[float]
        allows to specify the initial trust radius value for the trm
        relaxation algorithm
    symmetry_n_params: Optional[tuple[int, int]]
        Number of parameters for the lattice and fractional degrees of freedom
    symmetry_params: Optional[Sequence[str]]
        The list of all parametric constraints parameters
    symmetry_lv: Optional[tuple[str, str, str]]
        The list of parametric constraints for the lattice_vectors
    symmetry_frac: Optional[Sequence[tuple[str, str, str]]]
        The list of parametric constraints for the fractional coordinates
    symmetry_frac_change_threshold: Optional[float]
        Specifies the maximum allowed change in the initial structure to any
        fractional coordinate after applying the parametric constraints. If set to 1.0,
        then these tests will be ignored. max_change is the maximum allowed change in
        the fractional coordinates of an atom
    homogeneous_field: Optional[Vector3D]
        Allows performing a calculation for a system in a homogeneous electrical
        field E in units of V/AA.
    multipole: Optional[Sequence[tuple[float, float, float, int, float]]]
        Places the center of an electrostatic multipole field at a specified location,
        to simulate an embedding potential.
            x : x coordinate of the multipole.
            y : y coordinate of the multipole.
            z : z coordinate of the multipole.
            order : Integer number, which specifies the order of the multipole
                    (0 or 1 ≡ monopole or dipole).
            charge : Real number, specifies the charge associated with the multipole.
    esp_constraint: Optional[tuple[float, float, float] | tuple[float, float]]
        PBC only. Define the constraints for the fit of the ESP-charges for
        each atom. Depending on the chosen method (`esp_constraint` method in
        control.in). Method 1 needs three parameters (χ, J00, w) 3.167 and method 2
        needs two parameters (q0, β) 3.169 as defined above
    verbatim_writeout: Optional[bool]
        If True write the geometry.in file into the output file
    calculate_friction: Optional[bool]
        Calculate friction for this geometry

    """

    atoms: Sequence[FHIAimsAtom] = None
    lattice_vectors: Matrix3D[float] | None = None
    lattice_constraints: Vector3D[bool] | None = None
    species_dict: dict[str, SpeciesDefaults] | None = field(default_factory=dict)
    hessian_block: Sequence[tuple[int, int, Matrix3D]] | None = None
    hessian_block_lv: Sequence[tuple[float, float, Matrix3D]] | None = None
    hessian_block_lv_atoms: Sequence[tuple[float, float, Matrix3D]] | None = None
    hessian_file: bool | None = None
    trust_radius: float | None = None
    symmetry_n_params: tuple[int, int, int] | None = None
    symmetry_params: Sequence[str] | None = None
    symmetry_lv: tuple[str, str, str] | None = None
    symmetry_frac: Sequence[tuple[str, str, str]] | None = None
    symmetry_frac_change_threshold: float | None = None
    homogeneous_field: Vector3D | None = None
    multipole: Sequence[tuple[float, float, float, int, float]] | None = None
    esp_constraint: tuple[float, float, float] | tuple[float, float] | None = None
    verbatim_writeout: bool | None = None
    calculate_friction: bool | None = None

    def __post_init__(self):
        """Set up optional inputs given information/verify all inputs are correct."""
        if self.lattice_vectors is not None:
            self.lattice_vectors = np.array(self.lattice_vectors)
            if self.lattice_vectors.shape != (3, 3):
                raise InvalidGeometryError(
                    "Lattice vectors must be None or a 3x3 matrix"
                )

            if np.abs(np.linalg.det(self.lattice_vectors)) < 1e-12:
                raise InvalidGeometryError(
                    "Lattice vectors must be linearly independent"
                )

            if self.lattice_constraints is None:
                self.lattice_constraints = np.zeros((3, 3), dtype=bool)
        else:
            if self.lattice_constraints is not None:
                raise InvalidGeometryError(
                    "Lattice vectors must be defined for "
                    "lattice_constraints to be defined."
                )

            if any(
                inp is not None
                for inp in [
                    self.symmetry_n_params,
                    self.symmetry_params,
                    self.symmetry_lv,
                    self.symmetry_frac,
                    self.symmetry_frac_change_threshold,
                ]
            ):
                raise InvalidGeometryError(
                    "Lattice vectors must be defined when "
                    "using parametric constraints."
                )
        self.verify_object_lens()

    @classmethod
    def _get_property_names(cls, atomic=False):
        """Return a list of all properties defined for this geometry."""
        if atomic:
            return [
                k
                for k, v in cls.__dict__.items()
                if isinstance(v, property) and k not in ["ase_atoms", "structure"]
            ]
        # noinspection PyTypeChecker
        return [f.name for f in fields(cls)]

    def __repr__(self):
        """Return a representation of the geometry, featuring a simple chemical
        formula.
        """
        symbols = Counter(a.symbol for a in self.atoms)
        return (
            f"{self.__class__.__name__}"
            f"({''.join(f'{k}{v}' for k, v in symbols.items())})"
        )

    def __len__(self):
        """Get the length of the geometry."""
        return len(self.symbols)

    def verify_object_lens(self):
        """Verify the sizes of each input."""
        for item in ["lattice_constraints", "symmetry_lv"]:
            val = getattr(self, item, None)
            if val is not None and np.array(val).shape != (3, 3):
                raise InvalidGeometryError(
                    f"The shape of {item} must be 3x3 not {np.array(val).shape}"
                )

    @classmethod
    def from_file(cls, in_file: Path | str | TextIO) -> AimsGeometry:
        """Read in a geometry.in file and construct a Geometry object.

        Parameters
        ----------
        in_file: str | Path | TextIO
            File to load in

        Returns
        -------
        AimsGeometry
            The Geometry associated with the file

        """
        if isinstance(in_file, str | Path):
            with open(in_file) as fd:
                lines = fd.readlines()
        else:
            lines = in_file.readlines()

        return cls.from_strings(lines)

    @classmethod
    def from_strings(cls, lines: list[str]) -> AimsGeometry:
        """Read in a geometry.in file and construct a Geometry object.

        Parameters
        ----------
        lines: list[str]
            a list of all lines in the file.

        Returns
        -------
            The Geometry object associated with the file

        """
        lattice_vectors = []
        symbols: list[str] = []
        coords: list[Vector3D] = []

        is_empty: list[bool] = []
        is_pseudocore: list[bool] = []
        is_fractional: list[bool] = []

        hessian_block: None | list[tuple[int, int, Matrix3D]] = []
        hessian_block_lv: None | list[tuple[float, float, Matrix3D]] = []
        hessian_block_lv_atoms: None | list[tuple[float, float, Matrix3D]] = []

        symmetry_n_params: None | tuple[int, int, int] = (0, 0, 0)
        symmetry_params: None | Sequence[str] = []
        symmetry_lv: None | list[list[str]] = []
        symmetry_frac: None | list[list[str]] = []

        multipole: None | list[tuple[float, float, float, int, float]] = []

        symmetry_frac_change_threshold: None | float = None
        hessian_file: None | bool = None
        trust_radius: None | float = None
        homogeneous_field: Vector3D = None
        esp_constraint: None | tuple[float, ...] = None
        verbatim_writeout: None | bool = None
        calculate_friction: None | bool = None

        velocities: dict[int, Vector3D] = {}
        initial_charge: dict[int, float] = {}
        initial_moment: dict[int, float] = {}
        nuclear_constraints: dict[int, list[bool]] = {}
        lattice_constraints: None | dict[int, list[bool]] = {}
        constraint_regions: dict[int, int | None] = {}
        magnetic_response: dict[int, bool] = {}
        magnetic_moment: dict[int, float] = {}
        nuclear_spin: dict[int, float] = {}
        isotope: dict[int, float] = {}
        RT_TDDFT_initial_velocity: dict[int, Vector3D] = {}

        last_add: str = ""
        for line in lines:
            line = line.strip()
            if len(line) == 0 or line[0] == "#":
                continue
            inp = line.split("#")[0].split()

            if inp[0] in ["atom", "atom_frac", "empty", "pseudocore"]:
                symbols.append(inp[4])
                coords.append([float(ii) for ii in inp[1:4]])
                is_fractional.append(inp[0] == "atom_frac")
                is_empty.append(inp[0] == "empty")
                is_pseudocore.append(inp[0] == "pseudocore")
                last_add = "atom"
            elif inp[0] == "lattice_vector":
                lattice_vectors.append([float(ii) for ii in inp[1:4]])
                last_add = "lattice"
            elif inp[0] == "initial_moment":
                initial_moment[len(coords) - 1] = float(inp[1])
            elif inp[0] == "initial_charge":
                initial_charge[len(coords) - 1] = float(inp[1])
            elif inp[0] == "constrain_relaxation":
                if last_add == "atom":
                    nuclear_constraints[len(coords) - 1] = _create_constraints(inp)
                if last_add == "lattice":
                    lattice_constraints[len(coords) - 1] = _create_constraints(inp)
            elif inp[0] == "velocity":
                velocities[len(coords) - 1] = [float(ii) for ii in inp[1:4]]
            elif inp[0] == "RT_TDDFT_initial_velocity":
                RT_TDDFT_initial_velocity[len(coords) - 1] = [
                    float(ii) for ii in inp[1:4]
                ]
            elif inp[0] == "constraint_region":
                constraint_regions[len(coords) - 1] = int(inp[1])
            elif inp[0] == "magnetic_response":
                magnetic_response[len(coords) - 1] = True
            elif inp[0] == "magnetic_moment":
                magnetic_moment[len(coords) - 1] = float(inp[1])
            elif inp[0] == "nuclear_spin":
                nuclear_spin[len(coords) - 1] = float(inp[1])
            elif inp[0] == "isotope":
                isotope[len(coords) - 1] = float(inp[1])
            elif inp[0] == "symmetry_frac_change_threshold":
                symmetry_frac_change_threshold = float(inp[1])
            elif inp[0] == "hessian_file":
                hessian_file = True
            elif inp[0] == "trust_radius":
                trust_radius = float(inp[1])
            elif inp[0] == "homogeneous_field":
                homogeneous_field = np.array([float(e) for e in inp[1:]])
            elif inp[0] == "multipole":
                multipole.append(
                    (
                        float(inp[1]),
                        float(inp[2]),
                        float(inp[3]),
                        int(inp[4]),
                        float(inp[5]),
                    )
                )
            elif inp[0] == "esp_constraint":
                esp_constraint = tuple([float(ii) for ii in inp[1:]])
            elif inp[0] == "verbatim_writeout":
                verbatim_writeout = ".true." in inp[1]
            elif inp[0] == "calculate_friction":
                calculate_friction = ".true." in inp[1]
            elif inp[0] == "hessian_block":
                hessian_block.append(
                    (
                        int(inp[1]),
                        int(inp[2]),
                        np.array([float(ii) for ii in inp[3:12]]).reshape((3, 3)),
                    )
                )
            elif inp[0] == "hessian_block_lv":
                hessian_block_lv.append(
                    (
                        int(inp[1]),
                        int(inp[2]),
                        np.array([float(ii) for ii in inp[3:12]]).reshape((3, 3)),
                    )
                )
            elif inp[0] == "hessian_block_lv_atoms":
                hessian_block_lv_atoms.append(
                    (
                        int(inp[1]),
                        int(inp[2]),
                        np.array([float(ii) for ii in inp[3:12]]).reshape((3, 3)),
                    )
                )
            elif inp[0] == "symmetry_n_params":
                symmetry_n_params = (int(inp[1]), int(inp[2]), int(inp[3]))
            elif inp[0] == "symmetry_params":
                symmetry_params = inp[1:]
            elif inp[0] == "symmetry_lv":
                symmetry_lv.append([ii.strip() for ii in line[11:].strip().split(",")])
            elif inp[0] == "symmetry_frac":
                symmetry_frac.append(
                    [ii.strip() for ii in line[13:].strip().split(",")]
                )

        def dct2arr(
            dct: dict[int, Any], n_comp: int
        ) -> np.ndarray[Any, np.dtype[bool]] | None:
            """Convert a dictionary of atom indexes and values to an array."""
            if len(dct) == 0:
                return None
            values = list(dct.values())
            if isinstance(values[0], Sequence | np.ndarray):
                to_ret = np.zeros((n_comp, len(values[0])), dtype=type(values[0][0]))
            else:
                to_ret = np.zeros(n_comp, dtype=type(values[0]))
            for key, val in dct.items():
                to_ret[key] = val
            return to_ret

        if len(lattice_vectors) > 0:
            try:
                lattice_vectors = to_matrix3d(lattice_vectors)
            except AssertionError:
                raise InvalidGeometryError("There must be 3 lattice vectors")

            lattice_constraints = dct2arr(lattice_constraints, 3)

        for cc, coord in enumerate(coords):
            if is_fractional[cc]:
                coords[cc] = list(np.dot(coord, lattice_vectors))

        atoms = []
        for aa, sym in enumerate(symbols):
            atoms.append(
                FHIAimsAtom(
                    symbol=sym,
                    position=coords[aa],
                    velocity=velocities.get(aa),
                    initial_charge=initial_charge.get(aa, 0.0),
                    initial_moment=initial_moment.get(aa, 0.0),
                    constraints=nuclear_constraints.get(aa),
                    constraint_region=constraint_regions.get(aa),
                    magnetic_response=magnetic_response.get(aa),
                    magnetic_moment=magnetic_moment.get(aa),
                    nuclear_spin=nuclear_spin.get(aa),
                    isotope=isotope.get(aa),
                    is_empty=is_empty[aa],
                    is_pseudocore=is_pseudocore[aa],
                    RT_TDDFT_initial_velocity=RT_TDDFT_initial_velocity.get(aa),
                )
            )

        if len(lattice_vectors) > 0:
            for aa in range(len(atoms)):
                atoms[aa].set_fractional(lattice_vectors)

        if symmetry_n_params == (0, 0, 0):
            symmetry_lv = None
            symmetry_frac = None
            symmetry_params = None
            symmetry_n_params = None
            symmetry_frac_change_threshold = None

        if len(lattice_vectors) == 0:
            lattice_vectors = None
            lattice_constraints = None

        if len(hessian_block) == 0:
            hessian_block = None
        if len(hessian_block_lv) == 0:
            hessian_block_lv = None
        if len(hessian_block_lv_atoms) == 0:
            hessian_block_lv_atoms = None
        if len(multipole) == 0:
            multipole = None

        return cls(
            atoms=atoms,
            lattice_vectors=lattice_vectors,
            lattice_constraints=lattice_constraints,
            hessian_block=hessian_block,
            hessian_block_lv=hessian_block_lv,
            hessian_block_lv_atoms=hessian_block_lv_atoms,
            hessian_file=hessian_file,
            trust_radius=trust_radius,
            symmetry_n_params=symmetry_n_params,
            symmetry_params=symmetry_params,
            symmetry_lv=symmetry_lv,
            symmetry_frac=symmetry_frac,
            symmetry_frac_change_threshold=symmetry_frac_change_threshold,
            homogeneous_field=homogeneous_field,
            multipole=multipole,
            esp_constraint=esp_constraint,
            verbatim_writeout=verbatim_writeout,
            calculate_friction=calculate_friction,
        )

    @classmethod
    def from_atoms(
        cls,
        atoms,
        remove_velocities: bool = False,
        geo_constrain: bool = False,
        make_empty: list[int] | None = None,
        scaled: bool = False,
        wrap: bool = False,
    ) -> AimsGeometry:
        """Create an AimsGeometry from an ase.Atoms object.

        Parameters
        ----------
        atoms: ase.Atoms
            The atoms object to convert
        remove_velocities:bool
            If True don't add velocities to the geometry
        geo_constrain: bool
            If true add parameteric constraints
        make_empty: list[int]
            A list of indexes, to make an atom empty
        scaled: bool
            If True setup fractional coordinates
        wrap: bool
            If True and scaled is True wrap atoms into the unit cell

        """
        try:
            from ase.atoms import Atoms
            from ase.constraints import (
                FixAtoms,
                FixCartesian,
                FixCartesianParametricRelations,
                FixScaledParametricRelations,
            )
        except ModuleNotFoundError:
            raise TypeError("ase modules not found")

        if not isinstance(atoms, Atoms):
            raise TypeError(
                f"The atoms objectis not of type ase.Atoms, but f{type(atoms)}"
            )

        if scaled and not np.all(atoms.pbc):
            raise ValueError(
                "Requesting scaled for a calculation where scaled=True, but "
                "the system is not periodic"
            )

        if geo_constrain:
            if not scaled and np.all(atoms.pbc):
                warnings.warn(
                    "Setting scaled to True because a symmetry_block is detected.",
                    stacklevel=2,
                )
                scaled = True
            elif not np.all(atoms.pbc):
                warnings.warn(
                    "Parameteric constraints can only be used in periodic systems.",
                    stacklevel=2,
                )
                geo_constrain = False

        if (make_empty is not None) and (np.max(make_empty) >= len(atoms)):
                raise ValueError(
                    "The max value of make_empty (len(ghosts)) must not "
                    f"exceed the number of atoms {len(atoms)}"
                )

        geometry_props = {
            k: v
            for k, v in atoms.info.items()
            if k in AimsGeometry._get_property_names()
        }

        atomic_props = {
            k: v
            for k, v in atoms.info.items()
            if k in AimsGeometry._get_property_names(atomic=True)
        }

        symmetry_params = [[], []]
        symmetry_n_params = None
        symmetry_lv = None
        symmetry_frac = None
        cart_const = {}
        for const in atoms.constraints:
            if geo_constrain and isinstance(const, FixScaledParametricRelations):
                symmetry_params[1] = const.params
                symmetry_frac = const.expressions
            elif geo_constrain and isinstance(const, FixCartesianParametricRelations):
                symmetry_params[0] = const.params
                symmetry_lv = const.expressions
            elif isinstance(const, FixAtoms):
                cart_const.update(
                    dict.fromkeys(const.get_indices(), (True, True, True))
                )
            elif isinstance(const, FixCartesian):
                for ind in const.index:
                    cart_const[ind] = tuple(mm for mm in const.mask)

        if not remove_velocities:
            atomic_props["velocities"] = [
                v if any(v) else None for v in atoms.get_velocities()
            ]

        aims_atoms = []
        for i_at, atom in enumerate(atoms):
            atom_props = {singular(k): v[i_at] for k, v in atomic_props.items()}
            const = atom_props.pop("nuclear_constraint", None)
            aims_atoms.append(
                FHIAimsAtom(
                    symbol=atom.symbol,
                    position=atom.position,
                    initial_charge=atom.charge,
                    initial_moment=atom.magmom,
                    constraints=cart_const.get(i_at),
                    **atom_props,
                )
            )

        wrap = wrap and not geo_constrain
        if scaled:
            for atom in aims_atoms:
                atom.set_fractional(atoms.cell.array, wrap)

        if make_empty is not None:
            for aa in make_empty:
                aims_atoms[aa].is_empty = True

        if symmetry_lv is not None or symmetry_frac is not None:
            if symmetry_lv is None or symmetry_frac is None:
                raise ValueError("Both symmetry_lv and symmetry_frac must be defined")
            symmetry_n_params = (
                len(symmetry_params[0]),
                len(symmetry_params[1]),
                len(symmetry_params[0]) + len(symmetry_params[1]),
            )
            symmetry_params = symmetry_params[0] + symmetry_params[1]
        else:
            symmetry_params = None

        return cls(
            atoms=aims_atoms,
            lattice_vectors=None if atoms.cell.rank < 3 else atoms.cell,
            symmetry_n_params=symmetry_n_params,
            symmetry_params=symmetry_params,
            symmetry_lv=symmetry_lv,
            symmetry_frac=symmetry_frac,
            **geometry_props,
        )

    @classmethod
    def from_structure(cls, structure) -> AimsGeometry:
        """Create an AimsGeometry from an ase.Atoms object.

        Parameters
        ----------
        structure: Structure | Molecule
            The atoms object to convert

        """
        try:
            from pymatgen.core import Molecule, Species, Structure
        except ModuleNotFoundError:
            raise TypeError("pymatgen modules not found")

        if not isinstance(structure, Structure) and not isinstance(structure, Molecule):
            raise TypeError(
                f"The structure object is not a Structure or Molecule, "
                f"but type {type(structure)}"
            )

        lattice_vectors = getattr(structure, "lattice", None)
        if lattice_vectors is not None:
            lattice_vectors = lattice_vectors.matrix

        atoms = []
        for site in structure:
            element = site.species_string.split(",spin=")[0]
            charge = site.properties.get("charge", None)
            spin = site.properties.get("magmom", None)
            coord = site.coords
            v = site.properties.get("velocity", None)

            if isinstance(site.specie, Species) and site.specie.spin is not None:
                if spin is not None and spin != site.specie.spin:
                    raise ValueError(
                        "species.spin and magnetic moments don't agree. "
                        "Please only define one"
                    )
                spin = site.specie.spin
            elif spin is None:
                spin = 0.0

            if isinstance(site.specie, Species) and site.specie.oxi_state is not None:
                if charge is not None and charge != site.specie.oxi_state:
                    raise ValueError(
                        "species.oxi_state and charge don't agree. "
                        "Please only define one"
                    )
                charge = site.specie.oxi_state
            elif charge is None:
                charge = 0.0

            atoms.append(
                FHIAimsAtom(
                    symbol=element,
                    position=coord,
                    velocity=v,
                    initial_charge=charge,
                    initial_moment=spin,
                )
            )

        return cls(atoms=atoms, lattice_vectors=lattice_vectors)

    def load_species(self, species_directory: str | Path, overwrite: bool = False):
        """Create a species dictionary for the atoms.

        Parameters
        ----------
        species_directory:str | Path
            The Path to load the species into
        overwrite: bool
            True then overwrite existing species defaults

        """
        for sym in np.unique([at.symbol for at in self.atoms]):
            if sym in self.species_dict and not overwrite:
                warnings.warn(f"Won't overwrite species for {sym}", stacklevel=1)

            number = ATOMIC_SYMBOLS_TO_NUMBERS[sym]
            if number < 100:
                self.species_dict[sym] = SpeciesDefaults.from_file(
                    f"{species_directory}/{number:02d}_{sym}_default"
                )
            else:
                self.species_dict[sym] = SpeciesDefaults.from_file(
                    f"{species_directory}/{number:03d}_{sym}_default"
                )

    def set_species(self, sym: str, species: SpeciesDefaults):
        """Set a species default for a given symbol.

        Parameters
        ----------
        sym: str
            The symbol to add the species for
        species: SpeciesDefaults
            The species for the symbol

        """
        self.species_dict[sym] = species

    def get_species(self, sym: str) -> SpeciesDefaults:
        """Get a species default for a given symbol.

        Parameters
        ----------
        sym: str
            The symbol to add the species for

        Returns
        -------
            The species for the symbol

        """
        return self.species_dict[sym]

    def to_string(self):
        """Get the file content for this geometry."""
        content_str = []
        if self.lattice_vectors is not None and len(self.lattice_vectors) == 3:
            for lv, lv_const in zip(
                self.lattice_vectors, self.lattice_constraints, strict=False
            ):
                content_str.append(
                    f"lattice_vector {lv[0]:.15e} {lv[1]:.15e} {lv[2]:.15e}"
                )
                if np.any(lv_const):
                    content_str.append(
                        f"    constrain_relaxation "
                        f"{'x ' if lv_const[0] else ''}"
                        f"{'y ' if lv_const[1] else ''}"
                        f"{'z' if lv_const[2] else ''}"
                    )
        for atom in self.atoms:
            content_str.append(atom.to_string())

        if self.homogeneous_field is not None:
            if len(self.homogeneous_field) != 3:
                raise InvalidGeometryError(
                    "The provided homogeneous_field value is invalid"
                )
            content_str.append(
                f"homogeneous_field {self.homogeneous_field[0]:>20.12e} "
                f"{self.homogeneous_field[1]:>20.12e} "
                f"{self.homogeneous_field[2]:>20.12e}"
            )

        if self.multipole is not None:
            for mm, multipole in enumerate(self.multipole):
                if len(multipole) != 5:
                    raise InvalidGeometryError(
                        f"The provided {mm}th multipole value is invalid"
                    )
                content_str.append(
                    f"multipole {multipole[0]:>20.12e} {multipole[1]:>20.12e} "
                    f"{multipole[2]:>20.12e} {multipole[3]} "
                    f"{multipole[4]:>20.12e}"
                )
            # TODO: Specify a dipole moment to multipoles

        if self.esp_constraint is not None:
            if len(self.esp_constraint) not in (2, 3):
                raise InvalidGeometryError(
                    "The provided esp_constraint value is invalid"
                )
            content_str.append(
                f"esp_constraint "
                f"{' '.join(f'{x:>20.12e}' for x in self.esp_constraint)}"
            )

        if self.verbatim_writeout is not None:
            content_str.append(
                f"verbatim_writeout "
                f"{'.true.' if self.verbatim_writeout else '.false.'}"
            )

        if self.calculate_friction is not None:
            content_str.append(
                f"calculate_friction "
                f"{'.true.' if self.calculate_friction else '.false.'}"
            )

        if self.symmetry_n_params:
            content_str.append(
                f"symmetry_n_params "
                f"{self.symmetry_n_params[0]} {self.symmetry_n_params[1]} "
                f"{self.symmetry_n_params[2]}"
            )

        if self.symmetry_params:
            content_str.append("symmetry_params " + " ".join(self.symmetry_params))
        if self.symmetry_lv is not None:
            for line in self.symmetry_lv:
                content_str.append("symmetry_lv " + " , ".join(line))
        if self.symmetry_frac is not None:
            for line in self.symmetry_frac:
                content_str.append("symmetry_frac " + " , ".join(line))
        if self.symmetry_frac_change_threshold:
            content_str.append(
                f"symmetry_frac_change_threshold {self.symmetry_frac_change_threshold}"
            )

        if self.hessian_block is not None:
            for line in self.hessian_block:
                content_str.append(
                    f"hessian_block {line[0]} {line[1]} "
                    + " ".join([f"{ll:>20.12e}" for ll in np.array(line[2]).flatten()])
                )
        if self.hessian_block_lv is not None:
            for line in self.hessian_block_lv:
                content_str.append(
                    f"hessian_block_lv {line[0]} {line[1]} "
                    + " ".join([f"{ll:>20.12e}" for ll in np.array(line[2]).flatten()])
                )
        if self.hessian_block_lv_atoms is not None:
            for line in self.hessian_block_lv_atoms:
                content_str.append(
                    f"hessian_block_lv_atoms {line[0]} {line[1]} "
                    + " ".join([f"{ll:>20.12e}" for ll in np.array(line[2]).flatten()])
                )

        if self.hessian_file:
            content_str.append("hessian_file")
        if self.trust_radius is not None:
            content_str.append(f"trust_radius {self.trust_radius}")
        return "\n".join(content_str)

    def write_file(self, filename: str | Path | TextIO, header=None) -> None:
        """Write the geometry file."""
        if header is None:
            header = "\n".join(
                [
                    "#" + "=" * 79,
                    "# File written by pyfhiaims",
                    f"# {time.asctime()}",
                    "#" + "=" * 79,
                    "",
                ]
            )

        if isinstance(filename, str | Path):
            with open(filename, "w") as fd:
                content = header + self.to_string()
                fd.write(content)
        else:
            filename.write(header + self.to_string())

    @property
    def symbols(self):
        """The atomic symbol for each atom."""
        return [at.symbol for at in self.atoms]

    @property
    def numbers(self):
        """The atomic number for each atom."""
        return [at.number for at in self.atoms]

    @property
    def positions(self):
        """The position of each atom."""
        return [at.position for at in self.atoms]

    @property
    def fractional_positions(self):
        """The fractional positions of each atom."""
        return [at.fractional_position for at in self.atoms]

    @property
    def velocities(self):
        """The velocity for each atom."""
        return [at.velocity for at in self.atoms]

    @property
    def RT_TDDFT_initial_velocities(self):
        """The real-time TDDFT velocity for each atom."""
        return [at.RT_TDDFT_initial_velocity for at in self.atoms]

    @property
    def initial_charges(self):
        """The initial charge for each atom."""
        return [at.initial_charge for at in self.atoms]

    @property
    def initial_moments(self):
        """The initial moment for each atom."""
        return [at.initial_moment for at in self.atoms]

    @property
    def nuclear_constraints(self):
        """The nuclear constraints for each atom."""
        return [at.constraints for at in self.atoms]

    @property
    def constraint_regions(self):
        """The constraint region for earch atom."""
        return [at.constraint_region for at in self.atoms]

    @property
    def is_empty_atoms(self):
        """If each atom is an empty atom."""
        return [at.is_empty for at in self.atoms]

    @property
    def is_pseudocore_atoms(self):
        """If each atom is a pseudocore."""
        return [at.is_pseudocore for at in self.atoms]

    @property
    def magnetic_responses(self):
        """The magnetic reponse of all atoms."""
        return [at.magnetic_response for at in self.atoms]

    @property
    def magnetic_moments(self):
        """The magnetic moments of all atoms."""
        return [at.magnetic_moment for at in self.atoms]

    @property
    def nuclear_spins(self):
        """The nuclear spin of all atoms."""
        return [at.nuclear_spin for at in self.atoms]

    @property
    def isotopes(self):
        """The isotope of all atoms."""
        return [at.isotope for at in self.atoms]

    @property
    def species_block(self):
        """Get the species block for the control.in file."""
        # TODO: this does not belong here!
        if any(sym not in self.species_dict for sym in np.unique(self.symbols)):
            raise InvalidGeometryError(
                "Species are not defined for all atoms in the structure"
            )
        return "\n".join(
            [self.species_dict[sym].content for sym in np.unique(self.symbols)]
        )

    @property
    def n_atoms(self):
        """The number of atoms in the structure."""
        return len(self.symbols)

    @property
    def masses(self) -> np.ndarray:
        """Get the masses of the structure."""
        return self._species_property("mass")

    @property
    def nuclear_charges(self) -> np.ndarray:
        """Gets the nuclear charges of the structure."""
        return self._species_property("nucleus")

    @property
    def ase_atoms(self):
        """Get the ase.atoms object for the geometry."""
        try:
            from ase.atoms import Atoms
            from ase.constraints import (
                FixAtoms,
                FixCartesian,
                FixCartesianParametricRelations,
                FixScaledParametricRelations,
            )
        except ModuleNotFoundError:
            raise ModuleNotFoundError("ASE is not installed")

        info_entries = [
            p
            for p in self._get_property_names()
            if p
            not in (
                "structure",
                "ase_atoms",
                "atoms",
                "lattice_vectors",
                "species_dict",
            )
        ]

        info_entries += [
            p
            for p in self._get_property_names(atomic=True)
            if p
            not in (
                "symbols",
                "numbers",
                "positions",
                "fractional_positions",
                "velocities",
                "initial_charges",
                "initial_moments",
                "species_block",
                "n_atoms",
                "masses",
                "nuclear_charges",
            )
        ]

        cell = self.lattice_vectors if self.lattice_vectors is not None else (0, 0, 0)

        atoms = Atoms(
            numbers=self.numbers,
            positions=self.positions,
            velocities=[v if v is not None else [0, 0, 0] for v in self.velocities],
            magmoms=self.magnetic_moments,
            charges=self.initial_charges,
            pbc=self.lattice_vectors is not None,
            cell=cell,
            info={e: getattr(self, e) for e in info_entries},
        )

        fix_params = []
        if (self.symmetry_n_params is not None) and (
            np.sum(self.symmetry_n_params) > 0
        ):
            fix_params.append(
                FixCartesianParametricRelations.from_expressions(
                    list(range(3)),
                    self.symmetry_params[: self.symmetry_n_params[0]],
                    [expr for exprs in self.symmetry_lv for expr in exprs],
                    use_cell=True,
                )
            )

            fix_params.append(
                FixScaledParametricRelations.from_expressions(
                    list(range(len(atoms))),
                    self.symmetry_params[self.symmetry_n_params[0] :],
                    [expr for exprs in self.symmetry_frac for expr in exprs],
                )
            )

        fix_cart_const = []
        fixed_atoms = []
        for index, constraint in enumerate(self.nuclear_constraints):
            if constraint is None:
                continue

            if all(constraint):
                fixed_atoms.append(index)
            elif any(constraint):
                fix_cart_const.append(FixCartesian(index, constraint))

        if len(fixed_atoms) > 0:
            fix_cart_const.insert(0, FixAtoms(fixed_atoms))

        atoms.set_constraint(fix_cart_const + fix_params)

        return atoms

    @property
    def structure(self):
        """Get the structure with no properties."""
        return self.to_structure()

    def to_structure(self,
                     properties: dict[str, Any] | None = None,
                     site_properties: dict[str, Sequence[Any]] | None = None):
        """Get the pymatgen Structure or Molecule object for the geometry."""
        properties = properties or {}
        site_properties = site_properties or {}
        try:
            from pymatgen.core import Lattice, Molecule, Structure
        except ModuleNotFoundError:
            raise TypeError("pymatgen modules not found")

        charge = np.array([atom.initial_charge for atom in self.atoms])
        magmom = np.array([atom.initial_moment for atom in self.atoms])
        velocity: list[None | list[float]] = [atom.velocity for atom in self.atoms]
        species = [atom.symbol for atom in self.atoms]
        coords = [atom.position for atom in self.atoms]

        for vv, vel in enumerate(velocity):
            if vel is not None and np.sum(np.abs(vel)) < 1e-10:
                velocity[vv] = None

        lattice = (
            Lattice(self.lattice_vectors) if self.lattice_vectors is not None else None
        )

        site_props = {"charge": charge, "magmom": magmom}
        if any(vel is not None for vel in velocity):
            site_props["velocity"] = velocity

        site_props.update(**site_properties)
        if lattice is None:
            return Molecule(species, coords, np.sum(charge),
                            properties=properties, site_properties=site_props)

        return Structure(
            lattice,
            species,
            coords,
            np.sum(charge),
            coords_are_cartesian=True,
            properties=properties,
            site_properties=site_props,
        )

    def _species_property(self, prop: str) -> np.ndarray:
        """Return an array with property values corresponding to atomic Species."""
        if any(sym not in self.species_dict for sym in np.unique(self.symbols)):
            raise InvalidGeometryError(
                "Species are not defined for all atoms in the structure"
            )
        return np.array(
            [
                getattr(self.species_dict[sym], prop) * (not self.is_empty_atoms[ss])
                for ss, sym in enumerate(self.symbols)
            ]
        )


def _create_constraints(line: list[str]) -> list[bool]:
    """Set constraints."""
    constraints = [False, False, False]
    if "x" in line[1:] or "true" in line[1:]:
        constraints[0] = True
    if "y" in line[1:] or "true" in line[1:]:
        constraints[1] = True
    if "z" in line[1:] or "true" in line[1:]:
        constraints[2] = True
    return constraints


def singular(key: str) -> str:
    """Return the singular form of a string."""
    if key[-6:] == "_atoms":
        return key[:-6]
    if key[-4:] == "sses":
        return key[:-4] + "ss"
    if key[-3:] == "ies":
        return key[:-3] + "y"
    if key[-2:] == "es":
        return key[:-2] + "e"
    if key[-1:] == "s":
        return key[:-1]
    return key



"""A class representing a chunk of control.in file."""

from collections.abc import Sequence
from dataclasses import dataclass
from inspect import isclass

from monty.json import MSONable


@dataclass(kw_only=True)
class AimsControlChunk(MSONable):
    """Get the Aims Control Chunk."""

    keywords: tuple[str] = ()

    def __init_subclass__(cls, **kwargs):
        """Initialize the subclass."""
        super().__init_subclass__(**kwargs)
        keywords = []
        if not cls.keywords:
            for field in cls.__annotations__.values():
                if isclass(field) and issubclass(field, AimsControlChunk):
                    keywords += list(field.keywords)
            cls.keywords = tuple(keywords)

    @classmethod
    def from_strings(cls, config_str: Sequence[str]) -> "AimsControlChunk":
        """Get the Chunk from a string."""
        raise NotImplementedError

    def to_string(self) -> str:
        """Get the string of the chunk."""
        raise NotImplementedError

    def __str__(self) -> str:
        """Get the ControlChunk as a string."""
        return self.to_string()



"""Tests for AimsControl object."""

import numpy as np
import pytest

from pyfhiaims.control import AimsControl
from pyfhiaims.control.cube import AimsCube
from pyfhiaims.errors import PyaimsError
from pyfhiaims.geometry import AimsGeometry

control_in_content = """#
# List of parameters used to initialize the calculator:
#     output_level:MD_light
#     xc:pw-lda
#     relativistic:atomic_zora scalar
#     relax_geometry:trm 1E-2
#     relax_unit_cell:full
#     k_grid:2 2 2
#     output:[]
#===============================================================================
output_level                                      MD_light
xc                                                pw-lda
relativistic                                      atomic_zora scalar
relax_geometry                                    trm 1E-2
relax_unit_cell                                   full
k_grid                                            2 2 2
#===============================================================================


##################################################################################
#
# Read in species defaults for Al atom from pyfhiaims
#
##################################################################################
  species               Al
#     global species definitions
  nucleus               13.0
  mass                  26.9815386
#
  l_hartree             4
#
  cut_pot               3.5 1.5 1.0
  basis_dep_cutoff      0.0001
#
  radial_base    41   5.0000
  radial_multiplier        1
  angular_grids  specified
  division       0.6594  110
  division       0.8170  194
  division       0.9059  302
  outer_grid     302
##################################################################################
#
#  Definition of "minimal" basis
#
##################################################################################
#     valence basis states
    valence      3  s   2.
    valence      3  p   1.
#     ion occupancy
    ion_occ      3  s   1.
    ion_occ      2  p   6.
##################################################################################
#
#  Suggested additional basis functions. For production calculations,
#  uncomment them one after another (the most important basis functions are
#  listed first).
#
#  These were set using pyfhiaims, original may files have additional comments
#
##################################################################################
#  First tier
             ionic      3  d  auto
             ionic      3  p  auto
#            hydro      4  f  4.7
             ionic      3  s  auto
#  Second tier
#            hydro      5  g  7.0
#            hydro      3  d  6.0
#            hydro      2  s  11.6
#            hydro      2  p  0.9
#  Third tier
#            hydro      5  f  7.6
#            hydro      4  p  7.2
#            hydro      4  s  3.7
#            hydro      4  d  7.6
#  Fourth tier
#            hydro      4  d  13.6
#            hydro      5  g  11.2
#            hydro      4  d  0.9
#            hydro      1  s  0.4
#            hydro      4  p  0.1
#            hydro      5  f  9.8
#  Further basis functions
#            hydro      4  p  5.0
##################################################################################
#
# Read in species defaults for S atom from pyfhiaims
#
##################################################################################
  species               S
#     global species definitions
  nucleus               16.0
  mass                  32.065
#
  l_hartree             4
#
  cut_pot               3.5 1.5 1.0
  basis_dep_cutoff      0.0001
#
  radial_base    44   5.0000
  radial_multiplier        1
  angular_grids  specified
  division       0.4665  110
  division       0.5810  194
  division       0.7139  302
  outer_grid     302
##################################################################################
#
#  Definition of "minimal" basis
#
##################################################################################
#     valence basis states
    valence      3  s   2.
    valence      3  p   4.
#     ion occupancy
    ion_occ      3  s   1.
    ion_occ      3  p   3.
##################################################################################
#
#  Suggested additional basis functions. For production calculations,
#  uncomment them one after another (the most important basis functions are
#  listed first).
#
#  These were set using pyfhiaims, original may files have additional comments
#
##################################################################################
#  First tier
             ionic      3  d  auto
             hydro      2  p  1.8
#            hydro      4  f  7.0
             ionic      3  s  auto
#  Second tier
#            hydro      4  d  6.2
#            hydro      5  g  10.8
#            hydro      4  p  4.9
#            hydro      5  f  10.0
#            hydro      1  s  0.8
#  Third tier
#            hydro      3  d  3.9
#            hydro      3  d  2.7
#            hydro      5  g  12.0
#            hydro      4  p  10.4
#            hydro      5  f  12.4
#            hydro      2  s  1.9
#  Fourth tier
#            hydro      4  d  10.4
#            hydro      4  p  7.2
#            hydro      4  d  10.0
#            hydro      5  g  19.2
#            hydro      4  s  12.0
##################################################################################
#
# Read in species defaults for Zn atom from pyfhiaims
#
##################################################################################
  species               Zn
#     global species definitions
  nucleus               30.0
  mass                  65.409
#
  l_hartree             4
#
  cut_pot               3.5 1.5 1.0
  basis_dep_cutoff      0.0001
#
  radial_base    53   5.0000
  radial_multiplier        1
  angular_grids  specified
  division       0.5114   50
  division       0.8989  110
  division       1.2692  194
  division       1.6226  302
  outer_grid     302
##################################################################################
#
#  Definition of "minimal" basis
#
##################################################################################
#     valence basis states
    valence      4  s   2.
    valence      3  p   6.
    valence      3  d  10.
#     ion occupancy
    ion_occ      4  s   1.
    ion_occ      3  p   6.
    ion_occ      3  d   9.
##################################################################################
#
#  Suggested additional basis functions. For production calculations,
#  uncomment them one after another (the most important basis functions are
#  listed first).
#
#  These were set using pyfhiaims, original may files have additional comments
#
##################################################################################
#  First tier
             hydro      2  p  1.7
             hydro      3  s  2.9
             hydro      4  p  5.4
             hydro      4  f  7.8
             hydro      3  d  4.5
#  Second tier
#            hydro      5  g  10.8
#            hydro      2  p  2.4
#            hydro      3  s  6.2
#            hydro      3  d  3.0
#  Third tier
#            hydro      6  h  15.2
#            ionic      4  p  auto
#            hydro      5  s  12.8
#            hydro      4  f  5.4
#            hydro      4  d  7.0
#            hydro      4  f  20.0
#            hydro      3  p  2.2
#            hydro      5  f  6.4
#            hydro      5  g  8.0"""


def test_control(data_dir, tmp_path):
    control_dir = data_dir / "control"
    control_in = AimsControl.from_file(control_dir / "al2zns4.in")

    geometry_dir = data_dir / "geometry"
    geometry = AimsGeometry.from_file(geometry_dir / "scr.in")
    for key, val in control_in.species_defaults.items():
        geometry.set_species(key, val)

    content = control_in.get_content(
        geometry=geometry, verbose_header=True
    )

    assert control_in_content == content

    assert control_in["output_level"] == "MD_light"

    assert control_in["xc"] == "pw-lda"
    control_in["xc"] = "LDA"
    assert control_in["xc"] == "LDA"
    assert control_in.parameters["xc"] == "LDA"

    assert "Al" in control_in.species_defaults

    control_in["output"] = "test"
    assert control_in["output"] == []

    control_in.outputs.append("test")
    assert control_in["output"] == ["test"]

    del control_in["output"]
    assert control_in["output"] == []

    control_in.outputs.append("mulliken")
    control_in["cubes"] = [
        AimsCube(
            type="total_density",
            origin=[0.1, 0.2, -0.3],
            edges=np.eye(3) * 0.2,
        )
    ]
    control_in.write_file(
        geometry=geometry, writer=tmp_path, verbose_header=False, overwrite=True
    )

    new_control_in = AimsControl.from_dict(control_in.as_dict())
    for key, val in control_in.parameters.items():
        assert new_control_in[key] == val

    new_control_in["smearing"] = ("fermi-dirac", 0.1)
    assert (
        "occupation_type                                   fermi 0.1"
        in new_control_in.get_content(geometry)
    )

    new_control_in["smearing"] = ("methfessel-paxton", 0.1, 1)
    assert (
        "occupation_type                                   methfessel-paxton 0.100000 1"
        in new_control_in.get_content(geometry)
    )

    new_control_in["vdw_correction_hirshfeld"] = True
    new_control_in["relax_geometry"] = ("trm", 1e-4)
    new_control_in["charge"] = 1.0
    new_control_in["compute_forces"] = True
    new_control_in["compute_stress"] = None
    # new_control_in["species_dir"] = "./"

    assert "vdw_correction_hirshfeld" in new_control_in.get_content(
        geometry
    )
    assert (
        "relax_geometry                                    trm 0.0001"
        in new_control_in.get_content(geometry)
    )
    assert (
        "charge                                            1"
        in new_control_in.get_content(geometry)
    )
    assert (
        "compute_forces                                    .true."
        in new_control_in.get_content(geometry)
    )
    assert "compute_stress" not in new_control_in.get_content(geometry)

    new_control_in["xc"] = "libxc LDA_X+LDA_C_PW"
    assert "override_warning_libxc" in new_control_in.get_content(geometry)

    new_control_in["occupation_type"] = ("gaussian", 0.1)
    with pytest.raises(
        ValueError,
        match="Both smearing and occupation_type can"
    ):
        _ = new_control_in.get_content(geometry)

    with pytest.raises(
        ValueError,
        match="control.in file already in"
    ):
        control_in.write_file(
            geometry=geometry, writer=tmp_path, verbose_header=False, overwrite=False
        )

    del control_in["k_grid"]
    with pytest.raises(
        ValueError,
        match="k-grid must be defined for periodic"
    ):
        control_in.write_file(
            geometry=geometry, writer=tmp_path, verbose_header=False, overwrite=True
        )

    with pytest.raises(
        PyaimsError,
        match="Reading control.in from file does not work with cubes yet"
    ):
        _ = AimsControl.from_file(f"{tmp_path}/control.in")

    with pytest.raises(KeyError):
        _ = control_in["afeasfas"]
    # print(control_in.species_defaults["Co"].content)
